/*
* Compute own host name (domain name) in DNS dotted format, output 
* to 'name', with maximum length 'namelen'.
*
* If /etc/resolv.conf exists, the domain name is used, even if the DNS
* servers are not listed or available.  If the environment variable
* LOCALDOMAIN is defined, that value is used instead.  If
* /etc/resolv.conf does not exist and the local host is not a DNS
* server, a wrong domain name may be returned (at least on Solaris
* 2.4) as the value set with 'domainname' is used, minus the first
* word (e.g., if the domainname value is fokus.gmd.de, gmd.de is
* used).
*
* Additional sources of domain names:
*   /etc/defaultdomain (Solaris, SunOS)
*   sysinfo(SI_SRPC_DOMAIN)
*
* Copyright 1995 by Henning Schulzrinne
*/

#include <string.h>       /* strchr() */
#include <sys/types.h>
#include <netdb.h>        /* gethostbyname() */
#include <sys/utsname.h>  /* uname() */
#include <netinet/in.h>   /* struct sockaddr_in */
#include <arpa/nameser.h> /* MAXDNAME */
#include <arpa/inet.h>    /* inet_ntoa() */
#include <resolv.h>       /* res_init() */
#include <unistd.h>       /* getdomainname() */
#include <stdio.h>        /* sprintf() */
#include <ansi.h>         /* getdomainname() */


char *getinetdomain(char *name, int namelen)
{
  char *d = name;
  struct hostent *hep; /* gethostbyname */
  int domainlen;       /* characters left for domain */
  struct utsname un;
  extern int res_init(void);
  extern struct __res_state _res;

  /* get local host name (e.g., foo) */
  uname(&un);
  strcpy(name, un.nodename);
  domainlen = namelen - strlen(un.nodename) - 1;

  /* Check if that name already is a FQDN. */
  if (strchr(un.nodename, '.')) {
    return name;
  } 

  /* 
   * Try getdomainname() and check if it has a period.
   */
  d = name + strlen(un.nodename);
  if (getdomainname(d + 1, domainlen) == 0) {
    if (strchr(d + 1, '.')) {
      *d = '.';
      return name;
    }
  }

  /*
   * Try resolver.
   */
  res_init();
  if (_res.defdname[0] != '\0') {
    sprintf(name, "%s.%1.*s", un.nodename, domainlen, _res.defdname);
    return name;
  }
  else if (_res.dnsrch[0]) {
    sprintf(name, "%s.%*s", un.nodename, domainlen, _res.dnsrch[0]);
    return name;
  }

  /*
   * If this fails, try gethostbyname(). This name contains the
   * full domainname, including the host name.
   */
  hep = gethostbyname(un.nodename);
  if (hep) {
    if (strchr(hep->h_name, '.')) {
      strcpy(name, hep->h_name);
    }
    else {
      int i;
      for (i = 0; hep->h_aliases[i]; i++) {
        if (strchr(hep->h_aliases[i], '.')) {
          strcpy(name, hep->h_aliases[i]);
          break;
        }
      }
    }

    /* if none of the names are domain names, we use the IP address */
    if (strchr(name, '.') == NULL) {
      struct sockaddr_in sin;

      sin.sin_family = hep->h_addrtype;
      sin.sin_addr   = *(struct in_addr *)(hep->h_addr_list[0]);
      strcpy(name, inet_ntoa(sin.sin_addr));
    }
  }

  /* if gethostbyname() fails, get IP address through socket */

  return name;
} /* getinetdomain */
