#
# sip_response.tcl
# procedures to process an incoming SIP response
# - read the response
# - notify caller about response
# 



#
# get response from SIP server
# reads each line of an incoming TCP response and packs it into a list
# calls process_sip_response
#
proc read_sip_response {fd call_id} {
  global SIP msg response request invite

  if {![eof $fd]} {
    gets $fd line
    # check if a new response was received
    if {[lindex $line 0] == $SIP(VERSION)} {
      # get reply code and reason phrase
      set response($call_id,reply_code) [lindex $line 1]
      set response($call_id,reason_phrase) [lrange $line 2 end]

      # create new reply list
      set msg(payload_length) 0
      set msg($call_id) ""
      catch {unset msg($call_id,len)}
    }

    lappend msg($call_id) [string trimright $line]
    if {[info exists msg($call_id,len)]} {
      incr msg($call_id,len) -[string length $line]
      incr msg($call_id,len) -1
    } 

    # get payload length 
    if {[string match CONTENT-LENGTH:* [string toupper $line]] || [string match L:* [string toupper $line]]} {
      set msg(payload_length) [lindex [split [string trimright $line] :] 1]
    }
 
    if {[string match "\r" $line]} {
      # received an empty line, header complete
      set msg($call_id,len) $msg(payload_length)
    }

    # msg complete ?
    if {[info exists msg($call_id,len)]} {
      if {$msg($call_id,len) <= 0} {
        unset msg($call_id,len)
        process_sip_response $call_id $msg($call_id)
      }
    }
  } else {
    footer "Connection closed by server"
      after cancel $request($call_id,timer)
      if [info exist request($call_id,sock)] {
        close $fd
        catch {unset request($call_id,sock)}  
      }
      set index [lsearch $invite(active_invitations) $call_id]
      if {$index != -1} {
        set invite(active_invitations) [lreplace $invite(active_invitations) $index $index]
      }
    return 
  }
} ;# read_sip_response



#
# handler for incoming responses
# inform caller about incoming response
#
proc process_sip_response {call_id msg} {
  global response request image invite

  if [sip_extract_parameters $call_id response $msg] {
    return
  }

  if {$response($call_id,reply_code) >= 200} {
    # a definite response was received, kill timer
    after cancel $request($call_id,timer)
    if [info exists request($call_id,sock)] {
      close $request($call_id,sock)
      unset request($call_id,sock)
    } 
    set index [lsearch $invite(active_invitations) $call_id]
    if {$index != -1} {
      set invite(active_invitations) [lreplace $invite(active_invitations) $index $index]
    }
  }

  # create window to natofy caller about the received response
  set w .response[noperiod $call_id]
  if {![winfo exists $w]} {
      toplevel $w
      set x [expr [winfo screenwidth .] /2 ]
      set y [expr [winfo screenheight .] /2 ]
      wm geometry $w +$x+$y
    frame $w.head

    pack \
      [label $w.head.im] \
      [label $w.head.title -font *-*-*-bold-i-*-18-*-*-*-*-*-*-*] \
      -side left -anchor w 
    pack \
      $w.head \
      [label $w.l -text ""] \
      [button $w.b -text "Dismiss" -command "destroy $w"] \
      -side top
  }

  switch $response($call_id,reply_code) {
    100 {
      set title "Trying !"
      set image_name clock
      set text "The call of user $request($call_id,to) is send to the invitation server.\nThe invitation server is trying to locate the callee."
    }
    150 {
      set title "Ringing !"
      set image_name clock
      set text "A possible location of user \"$request($call_id,to)\" is found.\nThe invitation server is trying to alert the callee."
    }
    200 {
      # OK   
      # request accepted
      # create sessions
      sip_sound ok.au
      set title "Invitation Successful"
      wm title $w $title
      wm geometry $w +200+200
      $w.head.title configure -text $title
      $w.head.im configure -image $image(ok)
      $w.l configure -text "The call of user \"$request($call_id,to)\" was successful.\nThe appropriate sesssions will be created automatically.\nPlease wait!"
      lappend request(sessions_to_create) $call_id
      if {[llength $request(sessions_to_create)] == 1} {      
        sip_open_sessions
      }
      return
    }
    301 -
    302 -   
    603 {
      # Moved
      # create new buttons
      sip_sound afraid.au
      destroy $w.b
      frame $w.buttons
      pack \
        [button $w.buttons.m -text "Modify Request" -command "modify_sip_request $call_id"] \
        [button $w.buttons.b -text "Dismiss" -command "destroy $w"] \
        -side left
      pack $w.buttons
     
      switch $response($call_id,reply_code) {
        301 {
          set title "Moved Permanently !"
        }
        302 {   
          set title "Moved Temporarily !"
        }
        603 {
          set title "Alternative Address !"
        }
      }
      
      set image_name afraid
      set text "The call to \"$request($call_id,to)\" was unsuccessful because\nthe user or service is not available at this location.\nPlease try one or all of the following alternative locations:\n\n"    
     
      foreach loc $response($call_id,location) {
        append text "$loc\n\n"
      }
    }
    350 {
      # Alternative Service
      sip_sound afraid.au
      destroy $w.b
      frame $w.buttons
      pack \
        [button $w.buttons.m -text "Modify Request" -command "modify_sip_request $call_id"] \
        [button $w.buttons.b -text "Dismiss" -command "destroy $w"] \
        -side left
      pack $w.buttons
     
      set title "Alternative Service !"
      set image_name afraid
      set text "The call to \"$request($call_id,to)\" not successful,\nbut alternative services are possible\nPlease modify the invitation appropriately to the given parameters:\n\n"    
     
      append text "Contact-host: $response($call_id,host)\n"
      foreach line $response($call_id,payload) {
        append text "$loc\n"
      }
    }
    400 {
      sip_sound afraid.au
      set title "Bad Request !"
      set image_name afraid
      set text "The request to \"$request($call_id,to)\" could not\nbe understood due to malformed syntax"
    }
    401 {
      sip_sound afraid.au
      set title "Unauthorized !"
      set image_name afraid
      set text "The request to \"$request($call_id,to)\" requires user authentication"
    }  
    403 {
      sip_sound afraid.au
      set title "Forbidden !"
      set image_name afraid
      set text "The server understood the request to \"$request($call_id,to)\",\n but is refusing to fulfill it. "
    }
    404 {
      sip_sound afraid.au
      set title "Not Found !"
      set image_name afraid
      set text "The contacted server has definite information that the\nuser \"$request($call_id,to)\" does not exist \nin the specified domain."
    }
    406 {
      # Not Acceptable
      sip_sound afraid.au
      destroy $w.b
      frame $w.buttons
      pack \
        [button $w.buttons.m -text "Modify Request" -command "modify_sip_request $call_id"] \
        [button $w.buttons.b -text "Dismiss" -command "destroy $w"] \
        -side left
      pack $w.buttons

      set title "Not Acceptable"
      set image_name afraid
      set text "The invitation agent of user \"$request($call_id,to)\" was contacted\nsuccessfully but some aspects of the session profile were not acceptable.\nPlease modify and retransmit the request.\nReasons:\n\n"
      foreach r $response($call_id,reason) {
        switch $r {
          406.1 {
            append text "Insufficient Bandwidth\n\n"
          }
          406.2 {
            append text "Incompatible Protocol:\none or more protocols described in the request are not available\n\n"
          }
          406.3 {
            append text "Incompatible Format:\none or more media formats described in the request are not available\n\n"
          }
          406.4 {
            append text "Multicast not available\n\n"
          }
          406.5 {
            append text "Unicast not available\n\n" 
          }
        }
      }
    }
    450 {
      sip_sound afraid.au
      set title "Call Declined !"
      set image_name afraid
      set text "The user \"$request($call_id,to)\" was successfully contacted\nbut he explicitly does not wish to participate.\n Reason: \"$response($call_id,reason_phrase)\""
    }
    451 {
      sip_sound afraid.au
      set title "Busy !"
      set image_name afraid
      set text "The user's machine was successfully contacted,\nbut the user \"$request($call_id,to)\" was busy."
    }
    500 {
      sip_sound afraid.au
      set title "Sever Internal Error !"
      set image_name afraid
      set text "The invitation server encountered an unexpected condition that prevent\nit from fulfilling the request to \"$request($call_id,to)\"."
    }
    501 {
      sip_sound afraid.au
      set title "Not Implemented !"
      set image_name afraid
      set text "The invitation server does not support the functionality required\nto fulfill the request to \"$request($call_id,to)\"."
    } 
    503 {
      sip_sound afraid.au
      set title "Service Unavailable !"
      set image_name afraid
      set text "The invitation server is currently unable to handle the request to \"$request($call_id,to)\"."
    } 
    600 {
      sip_sound afraid.au
      set title "Search Failure !"
      set image_name afraid 
      set text "The invitation server understood the user's address,\nbut the request was unsuccessful in contacting the\nuser \"$request($call_id,to)\"."
    }
    601 {
      sip_sound afraid.au
      set title "Not Known Here !"
      set image_name afraid
      set text "The call to \"$request($call_id,to)\" was unsuccessful because\nthe user or service was not known at the address called."
    }
    602 {
      sip_sound afraid.au
      set title "Not Currently Here !"
      set image_name afraid
      set text "The call to \"$request($call_id,to)\" was unsuccessful because\nthe user or service is not currently located at the address called."
    }
    default {
      # reply codes which are not listed above are interpreted as x00 reply codes
      if ($response($call_id,reply_code) < 200) {
        set title "Trying !"
        set image_name clock
        set text "The call of user $request($call_id,to) is send to the invitation server.\nThe invitation server is trying to locate the callee."
      }
     
      if (($response($call_id,reply_code) > 200) && ($response($call_id,reply_code) < 300))  {
        # OK   
        # request accepted
        # create sessions
        sip_sound ok.au
        set title "Invitation Successful"
        wm title $w $title
	wm geometry $w +200+200
	$w.head.title configure -text $title
        $w.head.im configure -image $image(ok)
        $w.l configure -text "The call of user \"$request($call_id,to)\" was successful.\nThe appropriate sessions will be created automatically.\nPlease wait!"
        sip_open_sessions $call_id
        return
      }
      
      if (($response($call_id,reply_code) > 300) && ($response($call_id,reply_code) < 400))  {
        sip_sound afraid.au
        destroy $w.b
        frame $w.buttons
        pack \
          [button $w.buttons.m -text "Modify Request" -command "modify_sip_request $call_id"] \
          [button $w.buttons.b -text "Dismiss" -command "destroy $w"] \
           -side left
        pack $w.buttons
     
        set title "Alternative Address !"
        set image_name afraid
        set text "The call to \"$request($call_id,to)\" was unsuccessful because\nthe user or service is not available at this location.\nPlease try one or all of the following alternative locations:\n\n"    
     
        foreach loc $response($call_id,location) {
          append text "$loc\n\n"
        }
      }
  
      if (($response($call_id,reply_code) > 400) && ($response($call_id,reply_code) < 500))  {
        sip_sound afraid.au
        set title "Bad Request !"
        set image_name afraid
        set text "The request to \"$request($call_id,to)\" could not\nbe understood due to malformed syntax"
      }
     
      if (($response($call_id,reply_code) > 500) && ($response($call_id,reply_code) < 600))  {
        sip_sound afraid.au
        set title "Sever Internal Error !"
        set image_name afraid
        set text "The invitation server encountered an unexpected condition that prevent\nit from fulfilling the request to \"$request($call_id,to)\"."
      }

      if (($response($call_id,reply_code) > 600) && ($response($call_id,reply_code) < 700))  {
        sip_sound afraid.au
        set title "Search Failure !"
        set image_name afraid 
        set text "The invitation server understood the user's address,\nbut the request was unsuccessful in contacting the\nuser \"$request($call_id,to)\"."
      }
    }
  }
    
  wm title $w $title
  wm geometry $w +200+200
  $w.head.title configure -text $title
  $w.head.im configure -image $image($image_name)
  $w.l configure -text $text
} ;# process_sip_response



#
# get parameters of the request which corresponds the received response
#
proc get_original_invite_settings {call_id} {
  global request invite conf

  # get values specified for call_id
  set invite(guest) $request($call_id,to)
    
  if [info exists request($call_id,session_name)] {
    set conf(name) $request($call_id,session_name)
  }
    
  foreach tag {information uri phone email} {
    if [info exists request($call_id,$tag)] {
      set invite($tag) $request($call_id,$tag)
    }
  }
  set invite(session_addr) $request($call_id,session_addr)
  set invite(ttl) $request($call_id,ttl)
  
  set invite(media_list) $request($call_id,media_list)  
} ;# get_original_invite_settings



#
# set parameters corresponding to the incoming response
# call invite to let caller modify and retransmit the request
#
proc modify_sip_request {call_id} {
  global request invite response
    
  set w .response[noperiod $call_id]
  catch [destroy $w]
    
  # get invite settings specified for call_id
  get_original_invite_settings $call_id

  if {$response($call_id,reply_code) == 406} {
     set invite(media_list) $response($call_id,media_list)  
  }

  if {($response($call_id,reply_code) > 300) && ($response($call_id,reply_code) < 400) || $response($call_id,reply_code) == 603} {
    set invite(guest) [join $response($call_id,location)]
  }

  invite_proc
} ;# modify_sip_request



