/*
 *  udp.c --
 *
 *       UDP related functions (open connection, read and write, close)
 *
 *  (C) 1995 Bernd Deffner, GMD Fokus
 *      1996 Dorgham Sisalem
 *  See the copyright file for copyright information
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include "/usr/include/netinet/in.h"  /* struct sockaddr_in */
#include <arpa/inet.h>   /* inet_ntoa() */
#include <netdb.h>
#include <errno.h>
#include <values.h>
#include <unistd.h>      /* select(), perror() */
#include <string.h>
#include <iovec.h>
#include <tk.h>
#include "types.h"
#include "ansi.h"
#include "rtp.h"
#include "nevit.h"
#include "udp.h"
#include "debug.h"

static udp_t udp;
static char buf[MAX_PKT_LEN];

static char rcsid[]  = "$Id: udp.c,v 1.3 1998/04/06 08:24:36 dor Exp $";


/*
 *  Receive a packet
 */
int UDP_receive(ClientData client)
{
    int cli_len, len, proto, proto1;
    struct sockaddr_in cli_addr;
    proto = (int)client;
    proto1= (int)client;
    cli_len = sizeof(cli_addr);
    
   /* this is a ##hack## for some reason the tool looses contact to isc when
    * displaying its with a high frame rate.
    * Therefore we check the pmm socket before every recvfrom to see
    * if some messages were received. 
    */
    len = recvfrom(udp.sock[PMM],buf, MAX_PKT_LEN, 0,
		  (struct sockaddr *)&(s->rmt_addr[1]), &cli_len);
    /*
     * Linux 2.1.x Kernels do not provide a correct errno for select
     */

#ifdef LINUX
    if(len > 0) {  
#else
    if(errno != EWOULDBLOCK) {
#endif
	udp.proto_handler[PMM](udp.session, buf, len, (int )client);
    }
    if(proto==PMM) {
      return 1;
    }
    len = recvfrom(udp.sock[proto],buf, MAX_PKT_LEN, 0,
		  (struct sockaddr *)&(s->rmt_addr[1]), &cli_len);
    s->rmt_addr[1].sin_port=ntohs(s->rmt_addr[1].sin_port);
    if(proto>NUM_LEVELS)
      proto-=NUM_LEVELS;
    udp.proto_handler[proto1](udp.session, buf, len, (int )proto);
    return 1;
}


/*
 *  transmit a packet. Allowes to specify two buffers for scattered write
 *  buf2 must be NULL if only one buffer is necessary
 */
int UDP_write(int proto, char *buf1, int len1, char *buf2, int len2)
{
  struct msghdr msg;
  struct iovec iov[2];

  iov[0].iov_base = buf1;
  iov[0].iov_len = len1;
  iov[1].iov_base = buf2;
  iov[1].iov_len = len2;

  bzero(&msg,sizeof(struct msghdr));

  msg.msg_iov = iov;
  msg.msg_iovlen = 2;
  msg.msg_name = (caddr_t *) &udp.serv_addr[proto];
  msg.msg_namelen = sizeof(udp.serv_addr[proto]);

  sendmsg(udp.sock[proto], &msg, 0);
  return len1 + len2;
}


/* 
 * open unicast or multicast UDP connection
 *
 * session: pointer to global session information
 * sin_addr: socket address
 * proto: protocol (RTP, RTCP or conference bus protocol PMM)
 * ttl: time to live for multicast (ignored for unicast)
 * notify: create event handler for the socket descriptor
 * proto_handler: callback function if packet received on the socket
 *
 * To accomodate multi homed hosts we bind the opened sockets to the 
 * address of the active interface -this address is specified at isc-
 * For teh case of multicast we have to bind the sockets with the multicast
 * address but specify the output interface during the IP_ADD_MEMBERSHIP
 * this way the source field of the IP packets always have the right address
 * -in the usual case using just the default values leads the usage of wrong 
 * interface addresses
*/
static struct sockaddr_in* UDP_connect(char* ses,int proto,char *host,
				       char *local, int port, u_int8 ttl, 
				       int notify, void (*proto_handler)){
  int reuse = 1;  /* reuse address */
  struct sockaddr_in sin;
  int multicast;
  extern struct in_addr host2ip(char *);
  struct in_addr sin_addr;
  struct in_addr interface;

  sin_addr = host2ip(host);
  udp.serv_addr[proto].sin_port = htons(port);
  udp.serv_addr[proto].sin_family = AF_INET;
  udp.serv_addr[proto].sin_addr = sin_addr;
  multicast = IN_MULTICAST(htonl(udp.serv_addr[proto].sin_addr.s_addr));
  memset(&sin, 0, sizeof(sin));		      
  sin.sin_family = AF_INET;
  if(local!=NULL) { 
    if(multicast>0)
      sin.sin_addr.s_addr = sin_addr.s_addr;
    else
      sin.sin_addr = host2ip(local); /* bind address */
    interface = host2ip(local); /* local muticast interface address */
  }
  else /* single homed host so use default values */
    sin.sin_addr.s_addr = multicast ? sin_addr.s_addr : htonl(INADDR_ANY);

  sin.sin_port = htons(port);
		      
  /* Open a UDP socket */
  
  udp.sock[proto] = socket(PF_INET, SOCK_DGRAM, 0);
  if (udp.sock[proto] < 0) {
    perror("can not open socket");
    return(NULL);
  }

  if (setsockopt(udp.sock[proto], SOL_SOCKET, SO_REUSEADDR, (char *)&reuse,
		 sizeof(reuse)) == -1){
  	 perror("setsockopt: reuseaddr");
	return(NULL);
  }

  if (multicast) {
    struct ip_mreq mreq;
    int one = 1;
    u_int8 loop =1;
    int size, len;
    if(proto!=PMM) /* we need to read our own packets only for the local
		    * multicast case */
      loop=0;
    mreq.imr_multiaddr.s_addr = sin_addr.s_addr;
    if(local) { /* multihomed host */
      mreq.imr_interface.s_addr = interface.s_addr;
      if (setsockopt(udp.sock[proto], IPPROTO_IP, IP_MULTICAST_IF,
		   (char *)&interface, sizeof(interface)) < 0) {
	perror("IP_MULTICAST_IF");
	return(NULL);
      }
    }
    else
      mreq.imr_interface.s_addr = htonl(INADDR_ANY);

    if (setsockopt(udp.sock[proto], IPPROTO_IP, IP_ADD_MEMBERSHIP,
		   (char *)&mreq, sizeof(mreq)) < 0) {
      perror("IP_ADD_MEMBERSHIP");
      return(NULL);
    }

    if (setsockopt(udp.sock[proto], IPPROTO_IP, IP_MULTICAST_TTL, 
		   (char *)&ttl,sizeof(ttl)) < 0) {
      perror("IP_MULTICAST_TTL");
      return(NULL);
    }
    
    if (setsockopt(udp.sock[proto], SOL_SOCKET, SO_REUSEADDR, (char *)&one,
		   sizeof(one)) == -1) {
      perror("setsockopt: reuseaddr");
      return(NULL);
    }
  
  
    if (setsockopt(udp.sock[proto],IPPROTO_IP,
		   IP_MULTICAST_LOOP, (char *)&loop, sizeof(loop)) < 0) {
      perror("IP_MULTICAST_LOOP");
      return(NULL);
    }
    /*
     * Make sure the send/recv buffers are of a reasonable size
     * (this is a problem under Solaris 2.x) -- the default size is 0.
     */
    size = 65536;
    len = sizeof(int);
    getsockopt(udp.sock[proto], SOL_SOCKET, SO_SNDBUF, (char *)&size, &len);
    if (size < 65536) {
      size = 65536;
    }
    len = sizeof(int);
    setsockopt(udp.sock[proto], SOL_SOCKET, SO_SNDBUF, (char *)&size, 
	       sizeof(int));
    getsockopt(udp.sock[proto], SOL_SOCKET, SO_RCVBUF, (char *)&size, &len);
    if (size < 65536) {
      size = 65536;
    }
    setsockopt(udp.sock[proto],SOL_SOCKET,SO_RCVBUF,(char *)&size, sizeof(int));
    
    
  } /* if multicast */


  if (bind(udp.sock[proto], (struct sockaddr *)&sin, sizeof(sin)) < 0) {
      /*
       * Binding to multicast address fails if VJ's multicast mods have
       * not been installed, so we try again with INADDR_ANY.
       */
      if (errno == EADDRNOTAVAIL) {
        sin.sin_addr.s_addr = INADDR_ANY;
	if (bind(udp.sock[proto], (struct sockaddr *)&sin, sizeof(sin)) < 0){
	  perror("bind");
	  return(NULL);
	}
      }
    } /* if bind */
  udp.proto_handler[proto] = proto_handler;

  /* set the pmm socket as nonblocking to allow for continous checking of 
   * the socket
   */
  fcntl(udp.sock[proto],F_SETFL,O_NONBLOCK);
  /*if(proto==PMM) {
    fcntl(udp.sock[proto],F_SETFL,O_NONBLOCK);
  }*/

  /* create event handler for the socket */
  if (notify) {
#ifdef NEW
    udp.sock_handler[proto] = udp.sock[proto];
    /*#if defined(NEW) || defined(LINUX)
      udp.sock_handler[proto] = udp.sock[proto];*/
#else
    udp.sock_handler[proto] = Tcl_GetFile((ClientData) udp.sock[proto],
    TCL_UNIX_FD);
#endif
    Tcl_CreateFileHandler((int)udp.sock_handler[proto],TCL_READABLE,
			(Tcl_FileProc* ) UDP_receive, (ClientData) proto);
  }

  return &(udp.serv_addr[proto]);
}





/* 
*  open UDP connections for the session
*/ 
struct sockaddr_in* UDP_session(session_t *session, char *host,char *local)
{
  int port;
  int ttl = 16;
  char *s;
  struct sockaddr_in* rmt_addr;
  
  /* parse host description */
  s = strchr(host, '/');
  if (!s) 
    return NULL;
  else {
    *s = '\0';
    port = htons(atoi(s+1));
    s = strchr(s+1, '/');
    if (s) {
      ttl = atoi(s+1);
    }
  }

  /* open UDP connections for RTP and RTCP port */
  UDP_init(session);
  
  if((UDP_connect((char *)session, RTCP, host, local, port + 1, ttl, 1,
	      RTCP_receive))==NULL)
    return(NULL);

  if((rmt_addr=UDP_connect((char *)session, RTP, host, local, port, ttl, 1, 
			   rtp_handler))==NULL)
    return(NULL);
  return rmt_addr;
}

/*
 *  close sockets and remove file event handler
 */ 
static int  UDP_close(int proto)
{
  Tcl_DeleteFileHandler((ClientData)udp.sock[proto]); 
  Tcl_DeleteFileHandler((int)udp.sock_handler[proto]); 
  if(udp.sock[proto]>0)
    close(udp.sock[proto]);
  return 1;
}

/*
 *  Initialise UDP transmit mode
 */
void UDP_init(session_t *session)
{
  udp.session = session;     /* provide pointer to session structure */

  /* set generic function pointers to the UDP functions */
  session->net.read       = UDP_receive;
  session->net.write      = UDP_write;
  session->net.connect    = UDP_connect;
  session->net.close      = UDP_close;
  session->net.read_pmm   = UDP_receive;
  session->net.write_pmm  = UDP_write;
}
