#ifndef _Audio_h
#define _Audio_h
/*
* Audio library interface
* (Ubique Doors Server Interface)
*
* Callers must supply a callback routine (event handler), specified
*   through Audio_init().
*
* For Unix, the application must also provide two routines for
* establishing file input and timers:
*   notify_set_input_func()
*   notify_set_periodic_func()
*   Their syntax and semantics are described in the file <notify.h>.
* The Tcl library is used to provide a hash function.
* Thus, linking is:
*   your_application -lnevot -lnetutil -lgsm -ltcl -lnotify -lm
*
* Skeleton sequence of calls:
*   #include "Audio.h"
*   Audio_session_t s;
*   audio_descr_t audio;
*   Audio_sp_t sp;
*
*   Audio_init(my_event_handler);
*   sp.port = 3456;
*   audio.encoding = AE_DVI4;
*   (etc.)
*   Audio_SessionAlloc("session name", &sp, &s);
*   Audio_SessionSetSDES(s, SDES_CNAME, "foo@cs.bar.edu");
*   Audio_SessionSetTTL(s, 16);
*   Audio_SessionSetDuplex(s, Audio_net_priority);
*   Audio_SessionSetKey(s, "secret");
*   Audio_SessionSetAudio(s, &audio, 1, 0.02);
*   Audio_SessionOpen(st, "224.2.0.1");
*   Audio_SessionListen(st, 1);
*   Audio_SessionTalk(st, 1);
*   ...
*   Audio_exit();
*
*   void my_event_handler(Audio_session_t m, Audio_member_t m,
*     Audio_event_t e, ...)
*   {
*     switch(e) {
*     case Audio_Error:
*       break;
*     }
*   }
*
* Events for external users occur in this rough sequence (order of SDES
* events is unpredictable except that CNAME is first):
*   Audio_hello
*   Audio_cname
*   Audio_name     (once after new member arrives, and after info changes)
*   ...
*   Audio_receive  (when listen status changes)
*   ...
*   Audio_active   (when member stops/starts talking)
*   Audio_alive 0  (when member hasn't been heard from)
*   Audio_alive 1  (when member has been heard from again)
*   Audio_bye      (when member departs)
*
* Copyright (c) 1995 Ubique Corp.
* $Id: Audio.h,v 1.1.1.1 1997/11/12 17:16:23 dor Exp $
*/

#include "types.h"   /* u_int32, etc. */
#include "audio_descr.h"

#ifndef AUDIO_INTERNAL
#include "Audio_sp.h"
typedef long Audio_member_t, Audio_session_t;
#endif
enum {Audio_MemberLocal = 0};

/* full duplex, net-mutes-mike, mike-mutes-net */
typedef enum {
  Audio_duplex, Audio_net_priority, Audio_local_priority
} Audio_duplex_t;

/* SDES types (must be aligned with rtp.h) */
typedef enum {
  SDES_END    = 0,
  SDES_CNAME  = 1, 
  SDES_NAME   = 2,
  SDES_EMAIL  = 3,
  SDES_PHONE  = 4,
  SDES_LOC    = 5,
  SDES_TOOL   = 6,
  SDES_NOTE   = 7,
  SDES_PRIV   = 8,
  SDES_IMG    = 9, 
  SDES_DOOR   = 10,
  SDES_SOURCE = 11
} SDES_t;

/* 
 * Return values and error indicators in call to event handler. 
 * Codes marked with # are possible through event handler.
 */
typedef enum {
  Audio_OK = 0,           /* success */
  Audio_ERROR = -255,     /*   internal error, check Unix errno */
  Audio_EAUDIO_OPEN,      /* # cannot open audio device (fatal) */
  Audio_EAUDIO_ACCES,     /* # access denied */
  Audio_EAUDIO_INTR,      /* # system call interrupted */
  Audio_EAUDIO_IO,        /* # hangup or error during open */
  Audio_EAUDIO_BUSY,      /* # audio device busy */
  Audio_EAUDIO_NXIO,      /* # device does not exist */
  Audio_EAUDIOOUT,        /* # audio output (check errno) */
  Audio_EAUDIOIN,         /* # audio input (check errno) */
  Audio_ENETIN,           /* # network input (check errno) */
  Audio_ENETOUT,          /* # network output (check errno) */ 
  Audio_EDUP,             /* # session/member already exists */
  Audio_ERATE,            /* # cannot convert audio sampling rate */
  Audio_ECHANNELS,        /* # cannot convert audio channels */
  Audio_EENC,             /* # cannot convert audio encoding */
  Audio_EPERM,            /* # permission denied */
  /* Audio_SessionOpen() return values */
  Audio_ECNOMCAST,        /* not compiled for multicast */
  Audio_EHNOMCAST,        /* host/kernel does not support multicast */
  Audio_EMIXED,           /* cannot mix multicast and unicast addresses */
  Audio_ESELF,            /* cannot connect to self */
  Audio_ENOHOST,          /* no such host */
  Audio_EINVAL,           /* invalid argument */
  Audio_EPROTONOSUPPORT,  /* transport protocol not supported */
  Audio_EODDPORT,         /* port number must not be odd */
  /* others (header errors, etc.) */
  Audio_ENOSESSION,       /* no session to play audio to */
  Audio_ECANTREAD,        /* can't read audio file */
  Audio_ECANTWRITE,       /* can't write audio file */
  Audio_EICONFID,         /* invalid (vat) conference id */
  Audio_EIAUDIO,          /* invalid audio type */
  Audio_EISSRC,           /* invalid RTP ssrc */
  Audio_EIHX,             /* invalid header extension */
  Audio_EIPAD,            /* invalid padding */
  Audio_EICSRC,           /* invalid CSRC count */
  Audio_EITYPE,           /* invalid control packet type */
  /* from here on, PC-specific return values */
  Audio_EPC = -128
} Audio_errno_t;

/*
 * Routine provided by calling program.
 * Also called with SDES_t if SDES information changes.
 * Event handler, called for events and errors.
 */
typedef enum {
  /*
   * New SDES information is available for session 's', member 'm'.
   * Note that Audio_MemberGetSDES() returns NULL for an SDES item
   * until the corresponding SDES event has occurred. SDES events
   * occur only when the corresponding information changes. The order
   * of SDES events is unpredictable except that SDES_CNAME will be
   * first.
   */
  Audio_cname  = SDES_CNAME,  /* (s, m, _cname, char *cname) */
  Audio_name   = SDES_NAME,
  Audio_email  = SDES_EMAIL,
  Audio_phone  = SDES_PHONE,
  Audio_loc    = SDES_LOC,
  Audio_note   = SDES_NOTE,
  Audio_door   = SDES_DOOR,
  Audio_img    = SDES_IMG,
  Audio_tool   = SDES_TOOL,
  Audio_source = SDES_SOURCE,
  Audio_from   = 100, /* network source */
  /* 
   * Audio_receive through Audio_closed are asynchronous confirmations
   * of the Audio_Session...() commands.
   */
  /*
   * A session 's' has stopped/started receiving. (m = Audio_MemberLocal)
   * Invoked from within Audio_SessionListen().
   *  OR
   * A member 'm' in session 's' has stopped/started receiving.
   * Invoked from within Audio_MemberListen(). 
   */
  Audio_receive,     /* (s, 0, _receive, 0/1) */
  /*
   * A new session 's' has been created and parameters can be set now.
   */
  Audio_created,     /* (s, 0, _created) */
  /*
   * A new session 's' has been opened (network connectivity). 
   * Invoked from within Audio_SessionOpen().
   */
  Audio_opened,      /* (s, 0, _opened) */
  /*
   * An existing session 's' has been closed. Invoked from within
   * Audio_SessionDelete().
   */
  Audio_closed,      /* (s, 0, _closed)  */
  /*
   * The 'bye' event is generated when a member sends a 'bye'
   * control message (i.e., it departs from the sesssion).
   */
  Audio_bye,         /* member 'm' departed;    (s, m, _bye) */
  /*
   * A member is considered alive if we have received data or control
   * messages from it recently.  It is dead if not.  A member can be
   * declared 'dead' if the network is temporarily down, looses a
   * large number of packets or the remote member's application has
   * died without sending a 'bye' control message (or that message was
   * lost). An alive event with parameter 1 is generated only if
   * the member was previously declared dead (parameter=0).
   */
  Audio_alive,       /* (s, m, _alive, 0/1) */
   /*
    * A 'hello' event is generated when a new member arrives to
    * the session by sending a control message.
    */ 
  Audio_hello,       /* (s, m, _hello) */ 
   /*
	* An 'active event is generated when member 'm' starts
	* (parameter=1) or stops (parameter=0) talking.  The member 'm'
	* can be local or remote.
    */
  Audio_active,      /* (s, m, _active, 0/1) */ 
   /*
    * A 'vu' event is generated each time the VU meter value changes
    * for either the sending (parameter=0) or receiving side (1). The
    * level value is between 0 and 1.0.
    */
  Audio_vu,          /* (0, 0, _vu, 0/1, level) */
   /*
    * A 'vu_max' event is generated by the long-term maximum measurer.
    * It is typically used to generate a slowly-moving bar display as
    * part of a VU meter.
    */
  Audio_vu_max,      /* (0, 0, _vu, 0/1, level) */
   /*
    * Asynchronous error for member 'm' in session 's'.
    */
  Audio_error        /* (s, m, _error, Audio_errno_t) */
} Audio_event_t;

typedef	void (*Audio_handler_t)(Audio_session_t s, Audio_member_t m,
  Audio_event_t, ...);

/*
* Initialize audio engine; install call-back 'handler' for events.
*/
extern Audio_errno_t Audio_init(Audio_handler_t handler);

/*
* Initialize audio local address -only relevant to multihomed hosts.
*/
extern Audio_errno_t Audio_local(char* host);

/*
 * Close audio engine and clean up.
 */
extern void Audio_exit(void);

/***
 * Session routines.
 ***/
/*
 * Allocate session descriptor with name 'id' and properties 'sp'.  No
 * asynchronous events.  Call exactly once for each session.  Returns
 * Audio_OK on success, Audio_EDUP if the session identifier 'id'
 * exists already.
 */
extern Audio_errno_t Audio_SessionAlloc(char *id, Audio_sp_t *sp, 
  Audio_session_t *);

/* 
 * Add hosts to session 's'.  Generates Audio_created event.  Returns
 * Audio_OK on success, Audio_ECNOMULTI, Audio_EHNOMULTI,
 * Audio_EMIXED, Audio_ESELF, Audio_ENOHOST, Audio_EINVAL,
 * Audio_EPROTONOSUPPORT on failure.
 */
extern Audio_errno_t Audio_SessionOpen(Audio_session_t s, char *host);

/*
 * Enable/disable sending to session 's' (talking).  Opens audio
 * device when enabled, closes it if this is the last session with
 * talking enabled and there's no audio to be played from the network. 
 * Returns Audio_OK on success, Audio_EAUDIOBUSY if audio device is
 * busy (used by another application), Audio_EAUDIOOPEN if it could
 * not be opened for another reason (e.g., permission denied, no such
 * device; check errno for details).  No asynchronous events.
 */
extern Audio_errno_t Audio_SessionTalk(Audio_session_t s, const int flag);

/*
 * Enable/disable listening to session 's'.  Does not open audio
 * device until audio packets arrive from the network.  (At that time,
 * an asynchronous Audio_error event may be generated if the device
 * cannot be opened.) If listening is disabled, audio packets from the
 * network for this session are ignored.  Returns Audio_OK on success,
 * Audio_ERROR on error (indicates internal bug).
 */
extern Audio_errno_t Audio_SessionListen(Audio_session_t s, const int flag);

/*
 * Set local (outgoing) SDES values.  The 'data' is saved by the
 * routine.  Note: SDES_CNAME *must not* be changed.  Returns Audio_OK
 * on success, Audio_EINVAL if 'sdes' parameter is out of range.
 */
extern Audio_errno_t Audio_SessionSetSDES(Audio_session_t s,
  const SDES_t sdes, const char *data);

/*
* Set outgoing encryption key to 'key' and enable encryption.  A 'key'
* value of NULL disables encryption.  Return Audio_OK on success,
* Audio_EINVAL if key is invalid.
*/
extern Audio_errno_t Audio_SessionSetKey(Audio_session_t s,
  char *key);

/*
* Set multicast ttl (time-to-live) value. Ignored for unicast.
* Return Audio_OK on success, Audio_EINVAL if ttl value is out of range.
*/
extern Audio_errno_t Audio_SessionSetTTL(Audio_session_t s,
  int ttl);

/*
* Set duplex/echo suppression mode.
* Return Audio_OK on success, Audio_EINVAL if invalid parameter.
*/
extern Audio_errno_t Audio_SessionSetDuplex(Audio_session_t s,
  Audio_duplex_t d);

/*
* Set outgoing audio format (encoding, sampling rate, channels), the
* stereo channel incoming audio is played on (default: 1) and the
* number of seconds of audio per outgoing network packets.  If
* 'packetization' is zero, leave unchanged.  Does not enable audio
* device.  Return Audio_OK on success, Audio_EINVAL if invalid
* parameter, Audio_ERATE if invalid sampling rate, Audio_ECHANNELS if
* invalid channel count, Audio_EENC if invalid encoding.
*/
extern Audio_errno_t Audio_SessionSetAudio(Audio_session_t s,
  audio_descr_t *a, int audio_channel, double packetization);

/*
 * Find session based on text identifier (the one passed as argument
 * to Audio_SessionAlloc()). Return 0 if no match.
 */
extern Audio_session_t Audio_SessionFind(char *id);

/*
 * Return textual session identifier of session 's' (set by 'id'
 * parameter in Audio_SessionAlloc), NULL if none found.
 */
extern char *Audio_SessionId(Audio_session_t s);

/*
 * Remove session 's'. Generate Audio_closed event.
 * Return Audio_OK on success, Audio_ERROR on internal error (bug).
 */
extern Audio_errno_t Audio_SessionDelete(Audio_session_t s);

/*
 ****
 * Member routines.
 ****
 */

/*
* Enable/disable receiving from single session member 's' by flag: 0
* disables listening, 1 enables it. Return Audio_OK on success,
* Audio_ERROR on internal failure (bug).
*/
extern Audio_errno_t Audio_MemberListen(Audio_member_t m, const int flag);

/*
* Delete single session member.  Generate Audio_bye event for member
* 'm', with reason 'reason'.  Return Audio_OK on success, Audio_ERROR
* on internal failure (bug). 
*/
extern Audio_errno_t Audio_MemberDelete(Audio_member_t m, char *reason);

/*
 * Return SDES information of type 'sdes' for member 'm';
 * NULL if no value.
 */ 
extern char *Audio_MemberGetSDES(Audio_member_t m, const SDES_t sdes);

/*
* Return lowest available SDES information (CNAME, NAME, ...); useful
* for error messages.
*/
extern char *Audio_MemberId(Audio_member_t m);

/*
* Return printable version of error identifier 'e'.
*/
extern char *Audio_strerror(Audio_errno_t e);
#endif
