/*
agc  -- Automatic Gain Control

Copyright 1995 by Henning Schulzrinne; all rights reserved
*/

#define AGC_PRIVATE
#include "types.h"   /* size_t */
#include <math.h>
#include <tcl.h>
#include <string.h>  /* strcmp() */
#include <stdlib.h>  /* atoi(), malloc() */

typedef struct {
  int enable;        /* enable AGC */
  float f;           /* filter coefficient */
  float low, high;   /* permissible region (0..1) */
  float estimate;    /* current smoothed volume estimate */
} agc_t;

#include "agc.h"


static double const DB_RANGE = 60.;
/* absolute dB to 0..1 */
#define DB_ABS(x) (1. + (x)/DB_RANGE)

/*
* Configure AGC (enable, tc, low, high).
*/
char *agc_trace(ClientData client, Tcl_Interp *interp, 
  char *name1, char *name2, int flags)
{
  agc_t *a = (agc_t *)client;
  char *value = Tcl_GetVar2(interp, name1, name2, 0);
  double v;

  if (!value) return NULL;
  v = atof(value);
  if (strcmp(name2, "enable") == 0) {
    a->enable = v;
  }
  else if (strcmp(name2, "high") == 0) {
    a->high = DB_ABS(v);
  }
  else if (strcmp(name2, "low") == 0) {
    a->low  = DB_ABS(v);
  }
  /* Time constant in packets; smaller of filter values */
  else if (strcmp(name2, "tc") == 0) {
    a->f = 1 - exp(-1/v);
  }
  return NULL;
} /* agc_trace */


/*
* Allocate an AGC structure and initialize to zero.
*/
agc_t *agc_init(void)
{
  agc_t *a = (agc_t *)calloc(1, sizeof(agc_t));
  a->enable = 0;
  a->high   = 0.7;
  a->low    = 0.4;
  a->f      = 0.0625;
  return a;
} /* agc_init */


/*
* Automatic gain control.
* Return 0 if no change, +/-1 if change.
*/
int agc(agc_t *a, audio_stats_t *as)
{
  if (!a->enable) return 0;
  /*
   * Compute low-pass-filtered energy estimate.
   */
  a->estimate += (as->c[0].peak - a->estimate) * a->f;
  if (a->estimate > a->high) {
    return -1;
  }
  else if (a->estimate < a->low) {
    return 1;
  }
  return 0;
} /* agc */
