/*********************************************************************
 *                                                                   *
 *  Voice Terminal (VT)                                              *
 *  June, 1991                                                       *
 *                                                                   *
 *  Written at USC/Information Sciences Institute from an earlier    *
 *  version developed by Bolt Beranek and Newman Inc.                *
 *                                                                   *
 *  Copyright (c) 1991 University of Southern California.            *
 *  All rights reserved.                                             *
 *                                                                   *
 *  Redistribution and use in source and binary forms are permitted  *
 *  provided that the above copyright notice and this paragraph are  *
 *  duplicated in all such forms and that any documentation,         *
 *  advertising materials, and other materials related to such       *
 *  distribution and use acknowledge that the software was           *
 *  developed by the University of Southern California, Information  *
 *  Sciences Institute.  The name of the University may not be used  *
 *  to endorse or promote products derived from this software        *
 *  without specific prior written permission.                       *
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR     *
 *  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED   *
 *  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR       *
 *  PURPOSE.                                                         *
 *                                                                   *
 ********************************************************************/



/*
 * mix.c -- Mix streams of PCM samples.
 *
 * Synopsis:
 *      mix2(len, v0, v1);
 *      u_int8 *v0;   (Ptr to PCM buffer 0; sum is left there)
 *      u_int8 *v1;   (Ptr to PCM buffer 1)
 *
 * Description:
 *      These routines accept 2 buffers of u-Law PCM voice
 *      samples and mix them together to be played out as one signal.
 *      The logarithmic u-Law PCM samples are taken in pairs to index a
 *      64K table of sums.  The mixing assumes that most of the time the
 *      sum will not exceed the maximum dynamic range of the u-Law PCM
 *      output, so the sum is not attenuated.  Instead the sums in the
 *      table have been clipped at the maximum positive and negative
 *      values.  The rather gigantic table is used because there is
 *      plenty of memory available and not much CPU time.  The resulting
 *      mixed stream is left in buffer 0.  The for loops have been
 *      unravelled by a factor of 4 for speed.
 */

#include "types.h"
extern u_int8 sum_table[256*256];

/*
* Mix two PCM u-law streams.
*/
void mix2_pcmu(u_int8 *v0, u_int8 *v1, size_t len)
{
  register i;                             /* Loop index */
   
  for (i = len; i > 0; i--, v0++)
    *v0 = sum_table[(((*v0) << 8) | (*v1++))];
} /* mix2_pcmu */


/*
* Mix two streams encoded as L16 (16-bit linear) values. Does not check
* for overflow (will simply wrap around).
*/
void mix2_l16(int16 *v0, int16 *v1, size_t len)
{
  register i;

  for (i = len >> 1; i > 0; i--, v0++) *v0 += *v1++;
} /* mix2_l16 */


/*
* Mix two streams encoded as L8 (8-bit linear) values. Does not check
* for overflow (will simply wrap around).
*/
void mix2_l8(int8 *v0, int8 *v1, size_t len)
{
  register i;
  for (i = len; i > 0; i--, v0++) *v0 += *v1++;
} /* mix2_l8 */
