/*
* Allow only one copy of process per host and, possibly, user.  Thus,
* we use the pid file /usr/tmp/PROG-uid.pid if the 'user' flag is
* true, /usr/tmp/PROG if the user flag is false.  Return -1 if general
* error, 0 if success, >0 is pid of running daemon, < 0 if file
* exists, but process doesn't.
* 
* For reaons unknown, fcntl fails on SunOS 4.1.3 with EINVAL if
* /tmp/?.pid is used, but works if /usr/tmp/?.pid is used instead.
* This may have something to do with /tmp being used for swap as well.
*
* Based on W. Stevens, p. 376.
*/

#include <sys/types.h>
#include <sys/stat.h>  /* SEEK_ */
#include <signal.h>    /* kill() */
#include <errno.h>
#include <stdlib.h>    /* atoi() */
#include <string.h>    /* strlen() */
#include <fcntl.h>     /* F_ */
#include <unistd.h>    /* getuid() */
#include <stdio.h>     /* sprintf() */
#include "ansi.h"



/*
* Return pid if process exists, -pid if it doesn't, -1 if we couldn't
* read the file.
*/
static int readpid(int fd, char *name)
{
  int pid;
  char value[80];

  if (fd < 0) {
    if ((fd = open(name, O_RDONLY, 0)) < 0) return -1;
  }

  if (read(fd, value, sizeof(value)) < 0) return -1;
  pid = atoi(value);
  if (kill(pid, 0) < 0) {
    if (errno == ESRCH) return -pid;
    else return -1;
  }
  return pid;
} /* readpid */


int onlyone(char *prog, int user)
{
  int fd, val;
  char name[80];
  struct flock lock;

  if (user)
    sprintf(name, "/usr/tmp/%s-%d.pid", prog, (int)getuid());
  else
    sprintf(name, "/usr/tmp/%s.pid", prog);

  /* open file */
  if ((fd = open(name, O_RDWR | O_CREAT, 0)) < 0) {
    return readpid(fd, name);
  }

  /*
   * Since open() and's in the uname mask, we set the permission bits
   * explicitly to rw-rw-rw.
   */
  if (chmod(name,  
    S_IRUSR | S_IWUSR | S_IWGRP | S_IWOTH | S_IRGRP | S_IROTH) < 0) {
    return readpid(fd, name);
  }

  /* try and set a lock on the entire file */
  lock.l_type   = F_WRLCK;
  lock.l_start  = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len    = 0;  /* to EOF */
  if (fcntl(fd, F_SETLK, &lock) < 0) {
    if ((errno == EACCES) || (errno == EAGAIN)) return readpid(fd, name);
    return -1;
  }

  /* truncate to zero length, now that we have the lock */
  if (ftruncate(fd, 0) < 0) return -1;

  /* and write our process id */
  sprintf(name, "%d\n", (int)getpid());
  if (write(fd, name, strlen(name)) != strlen(name)) return -1;

  /* set close-on-exec for descriptor */
  if ((val = fcntl(fd, F_GETFD, 0)) < 0) return -1;

  val |= FD_CLOEXEC;
  if (fcntl(fd, F_SETFD, val) < 0) return -1;

  return 0;
} /* onlyone */
