/*
peak_meter  --  measure peak audio value

Copyright 1995 by Henning Schulzrinne; all rights reserved
*/

typedef struct {
  int enable;                    /* enable peakmeter */
  int interval;                  /* update every interval (measurements) */
  int interval_max;              /* reset interval */
  float hyst;                    /* hysteresis */
  unsigned long count;           /* counts seen */
  float old_value;               /* old display value */
  float max;                     /* current maximum */
  void (*show)(int,float);       /* function to show value */
  void (*show_max)(int, float);  /* function to show maximum bar */
  int arg;                       /* argument to 'show' routine */
} peakmeter_t;

#define PEAKMETER_PRIVATE
#include <stdlib.h>    /* calloc() */
#include <tcl.h>
#include <math.h>      /* fabs() */
#include <string.h>    /* strcmp() */
#include "peakmeter.h"
#include "pcmu_l.h"



/*
* Tcl trace handler for peakmeter (enable, interval, hyst).
*/
char *peakmeter_trace(ClientData client, Tcl_Interp *interp, 
  char *name1, char *name2, int flags)
{
  peakmeter_t *p = (peakmeter_t *)client;
  char *value = Tcl_GetVar2(interp, name1, name2, flags & TCL_GLOBAL_ONLY);
  int v;
  double const DB_RANGE = 60.;

  if (value != NULL) {
    v = atoi(value);
    if (strcmp(name2, "enable") == 0) {
      p->enable = v;
      if (p->show) (p->show)(p->arg, 0);
    }
    else if (strcmp(name2, "hysteresis") == 0) {
      p->hyst = atof(value) / DB_RANGE;
    }
    else if (strcmp(name2, "interval") == 0) {
      p->interval = v;
    }
  }
  return NULL;
} /* peakmeter_trace */


/*
* Allocated peakmeter datastructure.
*/
peakmeter_t *peakmeter_init(int arg, void (*show)(int, float), 
  void (*show_max)(int, float))
{
  peakmeter_t *p = (peakmeter_t *)calloc(1, sizeof(peakmeter_t));

  p->interval     = 5;     /* about 100 ms */
  p->interval_max = 250;   /* about 5 seconds */
  p->hyst         = 0.03;
  p->arg          = arg;
  p->show         = show;
  p->show_max     = show_max; 
  return p;
} /* peakmeter_init */


/*
* Peak meter with logarithmic display. Subtract DC estimate (as->avg).
* Values sent to display range from 0.0 to 1.0.
* There are several options for the longer-term peak meter:
* - reset every so often (apanel, vat); fewer updates
* - slow decay
*/
void peakmeter(peakmeter_t *p, audio_encoding_t encoding,
  audio_stats_t *as)
{
  float value, hyst;

  if (p->enable) {
    value = as->c[0].peak;
    if (++p->count % p->interval == 0) {
      hyst  = (value > 0) ? p->hyst : 0.;
      if (fabs(value - p->old_value) > hyst) {
        (p->show)(p->arg, value);
        p->old_value = value;
        as->c[0].peak = as->c[1].peak = 0;
      }
    }

    if (value > p->max) {
      p->max = value;
      (p->show_max)(p->arg, value);
    }
    else if (p->count % p->interval_max == 0) {
      p->max = value;
      (p->show_max)(p->arg, value);
    }
  }
} /* peakmeter */
