/*
rtp  --  RTP protocol handling

Copyright 1994-1995 by Henning Schulzrinne; all rights reserved
*/

#include "rtp.h"
#include "nevot.h"
#include "audio_generic.h"
#include "ring.h"
#include "member.h"
#include "bit.h"        /* ROUND() */
#include "util.h"       /* strcpy_pad(), timeval_ntp32(), PAD() */
#include "debug.h"      /* debug() */
#include <sys/types.h>
#include <sys/time.h>   /* gettimeofday() */
#include <stdlib.h>     /* free() */
#include <netinet/in.h>
#include <memory.h>
#include "ansi.h"



#define MAX_PT 128

audio_descr_t rtp_ad[MAX_PT] = {
/* enc sample ch */
  {AE_PCMU,  8000, 1},  /*  0 */
  {AE_MAX,   8000, 1},  /*  1 */
  {AE_G721,  8000, 1},  /*  2 */
  {AE_GSM,   8000, 1},  /*  3 */
  {AE_G723,  8000, 1},  /*  4 */
  {AE_DVI4,  8000, 1},  /*  5 */
  {AE_DVI4, 16000, 1},  /*  6 */
  {AE_LPC,   8000, 1},  /*  7 */
  {AE_MAX,      0, 1},  /*  8 (currently undefined) */
  {AE_MAX,      0, 1},  /*  9 (currently undefined) */
  {AE_L16,  44100, 2},  /* 10 */
  {AE_L16,  44100, 1},  /* 11 */
  /* rest is zero */
};

#define MAX_MISORDER 100
#define MAX_DROPOUT  3000

/*
* Update cycle count based on new sequence number.
*
* small positive delta (0 < sdelta < MAX_DROPOUT):
*   new < old: wrap
*   new > old: 
*   both: just update
* small negative delta (0 < -sdelta < MAX_MISORDER):
*   misordered: do nothing
* big |delta| [or MAX_DROPOUT < udelta < RTP_SEQ_MOD - MAX_MISORDER]
*   check if bad
*/
static int update_seq(sync_t *sync, stats_t *stats, u_int16 seq)
{
  u_int16 udelta = seq - (u_int16)sync->max_seq;

  /* in order, with permissible gap */
  if (udelta < MAX_DROPOUT) {
    if (seq < sync->max_seq) {
      sync->cycles += RTP_SEQ_MOD;
    }
    sync->max_seq = seq;
  }
  /* possibly restart */
  else if (udelta < RTP_SEQ_MOD - MAX_MISORDER) {
    if (seq == sync->bad_seq) {
      /*
       * two sequential packets -- assume that the other side
       * restarted without telling us so just re-sync
       * (i.e., pretend this was the first packet).
       */
    }
    else {
      sync->bad_seq = (seq + 1) & (RTP_SEQ_MOD-1);
      return 0;
    }
    sync->max_seq = seq;
  }
  else {
    stats->misorder++;
  }

  return 1;
} /* update_seq */


/*
* Return number of packets expected.
*/
static u_int32 rtp_packets_exp(Audio_member_t m)
{
  if (m->stats.received) {
    return m->sync.cycles + m->sync.max_seq - m->sync.first.seq + 1;
  }
  return 0;
} /* rtp_packets_exp */


/*
* Write single SDES item for a chunk, adjusting mbuf appropriately.
* Ignore if t.value is zero.
*/
static rtcp_sdes_item_t *rtcp_text(mbuf_t *mb, rtcp_sdes_item_t *rsp, 
  struct sdes *t)
{
  if (t->value && t->len && t->type) {
    rsp->type   = t->type;
    rsp->length = t->len;
    strcpy(rsp->data, t->value);
    mb->m_len += t->len + 2;
    rsp = (rtcp_sdes_item_t *)((char *)rsp + t->len + 2);
  }
  return rsp;
} /* rtcp_text */


/*
* Create SDES (source descriptor) field for member 's' with info 'e'
* (bit mask).  Return length (in words), not counting first word.
* Assume that common (32-bit) header is done elsewhere.
*/
static int rtp_write_sdes(mbuf_t *mb, Audio_member_t m)
{
  rtcp_sdes_item_t *rsp;
  src_id_t *src;
  struct sdes *sdes;
  int prior_len;       /* length of buffer prior to adding this SSRC */
  int pad;             /* bytes to pad */

  /* SSRC header */
  mb->m_len += 4;  /* skip generic header */
  prior_len = mb->m_len;
  src = mtod_end(mb, src_id_t *);
  *src = m->src;
  mb->m_len += 4;  /* skip SSRC */

  /* SDES items (CNAME, NAME, ...) for this SSRC */
  rsp = (rtcp_sdes_item_t *)(src + 1);
  for (sdes = m->sdes; sdes; sdes = sdes->next) {
    rsp = rtcp_text(mb, rsp, sdes);
  }
  /* terminate with zero byte */
  mtod_end(mb, char *)[0] = '\0';
  mb->m_len += 1;

  /* if necessary, pad with zeroes to next 4-byte boundary */
  pad = PAD(mb->m_len, 4);
  memset(mtod_end(mb, char *), RTCP_SDES_END, pad);
  mb->m_len += pad;

  /* length only for this particular SSRC */
  return ((mb->m_len - prior_len) >> 2);
} /* rtp_write_sdes */


/*
* Create SR or RR packet. If the station has neither sent nor
* received data, send empty RR.
*/
static void rtp_write_traffic(mbuf_t *mb, Audio_session_t s)
{
  Audio_member_t m;
  rtcp_rr_t *rr;
  int len = 1;
  rtcp_t *r;
  struct timeval now;
  u_int32 now_ntp32;

  /* sender report */
  r = mtod_end(mb, rtcp_t *);
  r->common.count = 0;
  rr = r->r.rr.rr;
  r->r.rr.ssrc = s->local->src;
  if (s->local->active > 0) {
    r->common.pt = RTCP_SR;
    s->local->active--;
    r->r.sr.rtp_ts    = htonl(tx_samples() + s->initial_ts);
    gettimeofday_ntp(&r->r.sr.ntp_sec);  /* returns two words */
    r->r.sr.ntp_sec   = htonl(r->r.sr.ntp_sec);
    r->r.sr.ntp_frac  = htonl(r->r.sr.ntp_frac);
    r->r.sr.psent     = htonl(s->packets[0]);
    r->r.sr.osent     = htonl(s->bytes);
    rr = r->r.sr.rr; 
    len += 5;
  }
  else {
    r->common.pt = RTCP_RR;
  }
  /* receiver reports */
  gettimeofday(&now, 0);
  now_ntp32 = timeval_ntp32(&now);

  ALL_MEMBERS(s) {
    if (m != s->local && m->active > 0) {
      int newly_expected, newly_received;

      /* update statistics */
      m->stats.expected = rtp_packets_exp(m);
      newly_expected = (m->stats.expected - m->stats.last.expected);
      newly_received = (m->stats.received - m->stats.last.received);

      m->active--;
      rr->ssrc     = m->src;
      if (newly_expected)
        rr->fraction = (double)(newly_expected - newly_received) / newly_expected * 256;
      else
        rr->fraction = 0;
      rr->last_seq = m->sync.cycles + m->sync.max_seq;
      rr->lost     = m->stats.expected - m->stats.received;
      rr->jitter   = m->stats.jitter;
      rr->lsr      = htonl((m->lsr_sec << 16) | (m->lsr_frac >> 16));
      /* force to zero if no SR received yet */
      if (m->last[2].tv_sec)
        rr->dlsr   = htonl(now_ntp32 - timeval_ntp32(&m->last[2]));
      else
        rr->dlsr   = 0;
      rr++;
      r->common.count++;
      len += 6;

      m->stats.last.expected = m->stats.expected;
      m->stats.last.received = m->stats.received;
    }
  }
  r->common.length = len;
  mb->m_len += (len+1) << 2;
} /* rtp_write_traffic */


/*
* Create a control packet. Add items to end. For BYE, add reason.
*/
static mbuf_t *rtp_write_control(Audio_session_t s, int type, char *reason)
{
  rtcp_t *r;
  int len;
  mbuf_t *mb;
  static struct {
    int type;
    rtcp_type_t pt;  
  } list[] = {
    {P_traffic, RTCP_RR},
    {P_sdes,    RTCP_SDES}, 
    {P_bye,     RTCP_BYE}, 
    {0,0}
  };
  int i;

  mb = m_get("rwc");

  /* Add random prefix if encrypted. */
  if (s->encrypt) {
    mtod(mb, u_int32 *)[0] = random32(0);
    mb->m_len += 4;
  }

  for (i = 0; list[i].type; i++) {
    if (type & list[i].type) {
      r = mtod_end(mb, rtcp_t *);
      len = 0;
      r->common.pt      = list[i].pt;
      r->common.version = RTP_VERSION;
      r->common.p       = 0;
      switch(list[i].type) {
      case P_traffic:
        rtp_write_traffic(mb, s);
        break;

      case P_sdes:
        /* XXX if bridge, report on all sources */
        r->common.length = rtp_write_sdes(mb, s->local);
        r->common.count  = 1;
        break;

      case P_bye:
        /* XXX if bridge, report on all sources */
        r->r.bye.src[0] = s->local->src;
        len = 8 + strcpy_pad((char *)&r->r.bye.src[1], reason);
        r->common.count  = 1;
        r->common.length = (len >> 2) - 1;
        mb->m_len += len;
        break;
      }
    }
  }

  /* Add padding to last unit if encryption and necessary. */
  if (s->encrypt) {
    int pad = PAD(mb->m_len, 8); 
    if (pad) {
      mb->m_len += pad;
      mtod(mb, u_int8 *)[mb->m_len - 1] = pad;
      r->common.p = 1;
      r->common.length += (pad >> 2);
    }
  }
  return mb;
} /* rtp_write_control */


/*
* Write RTP control header at the beginning of the first audio buffer;
* attach remaining audio buffers. 'type' contains P_bot. Return mbuf.
*/
static mbuf_t *rtp_write_data(Audio_session_t s, ring_t *r, int type, 
  mbuf_t *mb)
{
  rtp_hdr_t *h;
  int i;
  int pad = 0;
  int csrc = 0;

  /* Count CSRC. */
  if (r) {
    for (i = 0; i < MAX_TALKING; i++) {
      if (r->member[i]) csrc++;
    }
  }

  /* Make room at front of buffer for header. */
  mprepend(mb, 12 + csrc * 4);

  /* if encryption, pad last buffer to next 8-byte boundary */
  if (s->encrypt) {
    mbuf_t *m, *m_last;

    for (m = mb; m; m = mb->m_next) {
      pad += mb->m_len;
      m_last = m;
    }
    pad = (8 - (pad & 0x7)) & 0x7;

    mtod(m_last, char *)[m_last->m_len + pad - 1] = pad;
    m_last->m_len += pad;
  }

  h = mtod(mb, rtp_hdr_t *);
  h->version = RTP_VERSION;
  h->p       = pad > 0;
  h->x       = 0;
  h->cc      = csrc;
  h->m       = (type & P_bot) != 0;
  h->pt      = s->rt.audio_header;
  h->seq     = htons(s->seq);
  h->ts      = htonl(mb->m_timestamp + s->initial_ts);
  h->ssrc    = s->local->src;

  /* insert CSRC list into RTP header */
  if (r) {
    for (i = 0; i < MAX_TALKING; i++) {
      if (r->member[i]) h->csrc[i] = r->member[i]->src;
    }
  }

  return mb;
} /* rtp_write_data */


/*
* Decode incoming SDES option. Return 0 if invalid type or length.
*/
static rtcp_sdes_t *rtp_read_sdes(Audio_member_t m,
  rtcp_sdes_item_t *rsp)
{
  for (; rsp->type; rsp = (rtcp_sdes_item_t *)((char *)rsp + rsp->length + 2)) {
    member_sdes(m, rsp->type, rsp->data, rsp->length);
  }
  return (rtcp_sdes_t *)ROUND(rsp, 4);
} /* rtp_read_sdes */


/*
* Decode incoming SR/RR report.
*/
static void rtp_read_traffic(Audio_session_t s, rtcp_t *r, tsap_t *nsrc)
{
  Audio_member_t m;
  int i;
  rtcp_rr_t *rr;

  /*
   * Find sender of SR/RR packet and set up network connection if
   * necessary.
   */
  m = member_find(s, MEMBER_CONTROL, nsrc, r->r.sr.ssrc);

  /* sender report */
  if (r->common.pt == RTCP_SR) {
    m->lsr_sec  = ntohl(r->r.sr.ntp_sec);
    m->lsr_frac = ntohl(r->r.sr.ntp_frac);
    m->last[2]  = m->last[1];
    s->senders++;
    rr = r->r.sr.rr;
  }
  else {
    rr = r->r.rr.rr;
  }

  /* process receiver reports */
  for (i = 0; i < r->common.count; i++) {
    if (rr[i].ssrc == s->local->src) {
      /* store information about us as seen by remote site */
    }
  }
} /* rtp_read_traffic */

/*
* Process RTCP packets. Return pointer to member record or zero, if invalid.
*/
Audio_errno_t rtp_read_control(Audio_session_t s, mbuf_t *mb,
  tsap_t *nsrc, Audio_member_t *mp)
{
  long len;  /* remaining length of combined RTCP packets in words */
  Audio_member_t m = (Audio_member_t)Audio_MemberLocal;
  rtcp_t *r;
  rtcp_sdes_t *rsi;
  char reason[256] = "";
  int i;

  len = mb->m_len >> 2;
  r = mtod(mb, rtcp_t *);
  while (len > 0) {
    len -= r->common.length + 1;
    if (len < 0 || r->common.version != RTP_VERSION) {
      /* something wrong with packet format; don't look at this */
      debug(DEBUG_CNTL, "rtp bad header: len=%d version=%d", 
        len, r->common.version);
      return Audio_EITYPE;
    }
    switch(r->common.pt) {
    case 3:  /* backward compatibility with old RTP version */
    case RTCP_BYE:
      if (r->common.length > r->common.count) {
        char *s;
        s = (char *)&r->r.bye.src[r->common.count];
        strncpy(reason, s+1, *s);
        reason[(int)*s] = '\0';
      }
      for (i = 0; i < r->common.count; i++) {
        m = member_find(s, MEMBER_DONT_ADD | MEMBER_CONTROL, nsrc, 
          ntohl(r->r.bye.src[i]));
        Audio_MemberDelete(m, reason);
      }
      break;

    case 2:  /* backward compatibility with old RTP version */
    case RTCP_SDES:
      for (i = 0, rsi = &r->r.sdes; i < r->common.count; i++) {
        /*
		 * Since the SDES may not be for an SSRC, but rather from a
		 * mixer, we can't open a network connection for the new
		 * member here.
         */
        m = member_find(s, MEMBER_CONTROL, nsrc, rsi->src);
        rsi = rtp_read_sdes(m, rsi->item);
      }
      break;

    case RTCP_SR:
    case RTCP_RR:
      rtp_read_traffic(s, r, nsrc);
      break;

	default:
      break;
    }
    r = (rtcp_t *)((u_int32 *)r + r->common.length + 1);
  }
  if (m) *mp = m;
  return Audio_OK;
} /* rtp_read_control */

/*
* Locate synchronization source. Remove header in mbuf.
* Translate CSRC list to list of pointers, adding SSRC as first.
* Return 0 if header error (e.g., wrong encryption key). Don't
* add member; only RTCP does this.
*/
Audio_errno_t rtp_read_data(Audio_session_t s, mbuf_t *mb, tsap_t *nsrc,
  int *cc, Audio_member_t *slist, Audio_member_t *mp)
{
  int i;
  Audio_member_t m = 0;
  rtp_hdr_t *h = mtod(mb, rtp_hdr_t *);

  m = member_find(s, MEMBER_DONT_ADD, nsrc, h->ssrc);
  if (!m) return Audio_EISSRC;

  /* 
   * Remove header extension; length is in second short of the word
   * after the fixed header.
   */
  if (h->x) {
    unsigned len = ntohs(((u_int32 *)h)[3 + h->cc]) & 0xffff;
    if (len < mb->m_len)
      m_adj(mb, len);
    else
      return Audio_EIHX;  /* bad packet */
  }

  m_adj(mb, 12 + h->cc * 4);

  /* decode audio format if it has changed (or on first packet) */
  if (h->pt != m->audio_header) {
    m->audio_header = h->pt;
    m->audio        = rtp_ad[m->audio_header];
    audio_out(m, &rtp_ad[m->audio_header]);
  }
  m->talkspurt |= h->m;

  /* time stamp and sequence number */
  m->sync.this.ts  = ntohl(h->ts);
  m->sync.this.seq = ntohs(h->seq);
  update_seq(&m->sync, &m->stats, m->sync.this.seq);

  /* CSRC list; add only those we already know about */
  *cc = 0;
  for (i = 0; i < (h->cc & 0x0f); i++) {
    slist[*cc] = member_find(s, MEMBER_DONT_ADD, nsrc, h->csrc[i]);
    if (slist[*cc]) *cc += 1;
  }
  /* include SSRC for convenience */
  slist[0] = m;
  *cc += 1;

  /* decrease length of payload by padding, if present */
  if (h->p) {
    mb->m_len -= mtod(mb, u_int8 *)[mb->m_len-1];
    if (mb->m_len < 0) return Audio_EIPAD;
  }

  *mp = m;
  return Audio_OK;
} /* rtp_read_data */


/*
* Change audio header for outgoing packets. Return -1
* if no match, 0 if ok.
*/
int rtp_change_audio(rt_t *rt, audio_descr_t *ad)
{
  int i;

  /* search for first matching audio descriptor */
  for (i = 0; i < MAX_PT; i++) {
    if (audio_cmp(ad, &rtp_ad[i])) {
      rt->audio_header = i;
      break;
    }
  }
  if (i == MAX_PT) return -1;

  return 0;
} /* rtp_change_audio */

/*
* Initialize a session for using RTP and outgoing audio encoding. 
* Must be called once (possibly with rt == 0) if incoming RTP packets
* are to be handled.  Create mapping from audio_descr to outgoing
* audio byte.  Return -1 if illegal encoding, 0 otherwise.
*/
void rtp_init(rt_t *rt)
{
  int i;

  if (rt) {
    rt->write_control = rtp_write_control;
    rt->write_data    = rtp_write_data;
    rt->packets_exp   = rtp_packets_exp;
    rt->change_audio  = rtp_change_audio;
  }
  /* initialize rtp_ad[] */
  for (i = 0; i < MAX_PT; i++) {
    if (rtp_ad[i].channels == 0) {
      rtp_ad[i].encoding = AE_MAX;
      rtp_ad[i].channels = 1;
    }
  }
} /* rtp_init */
