/*
--  maintain list of talkers

We maintain an unordered array of all currently talking members.
At each playout instant, we check whether the audio for that
member has expired. If so, we remove the pointer from the list and
signal via event(). For a continuous talker, the audio
expiration time will always be updated.

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/



#include "nevot.h"


/*
* Set talking indicator for a session member, starting immediately,
* regardless of for what time in the future the audio was actually
* inserted.
*/
void talking_set(ring_t *r, Audio_member_t m)
{
  int i;

  /* 
   * If not currently in talker list, add it.
   */
  if (!m->talking && m->id) {
    for (i = 0; i < MAX_TALKING; i++) {
      if (!r->member[i]) {
        m->talking = TRUE;
        r->member[i] = m;
        if (i > r->member_last) r->member_last = i;
        (*nevot.event)(m->s, m, Audio_active, TRUE);
        break;
      }
    }
  }
} /* talking_set */

/*
* Remove any members from active list that are no longer talking.
* XXX: We could add r->check_period to avoid updating the
* display for every playout instant.
*/
void talking_check(ring_t *r)
{
  int i;
  Audio_member_t m;

  for (i = 0; i <= r->member_last; i++) {
    m = r->member[i];
    if (m && (m->last_sample < r->next)) {
      (*nevot.event)(m->s, m, Audio_active, FALSE);
      m->talking   = FALSE;
      r->member[i] = 0;
      if (i >= r->member_last) r->member_last = i-1;
    }
  }
} /* talking_check */

/*
* Remove member with identifier 'm' from all talking indicator slots.
* (This avoids stale references in talking_show().)
*/
void talking_delete(Audio_member_t m)
{
  ring_t *r;
  int i;

  r = m->s->ring;
  for (i = 0; i <= r->member_last; i++) {
    if (m == r->member[i]) {
      (*nevot.event)(m->s, m, Audio_active, FALSE);
      r->member[i] = 0;
      if (i >= r->member_last) r->member_last = i-1;
    }
  }
} /* talking_delete */


/*
* Clear all left-over talking indications.
*/
void talking_clear(Audio_session_t s)
{
  ring_t *r = s->ring;
  int i;

  for (i = 0; i <= r->member_last; i++) {
    if (r->member[i]) {
      (*nevot.event)(r->member[i]->s, r->member[i], Audio_active, FALSE);
      r->member[i]->talking = FALSE;
      r->member[i] = 0;
    }
  }
} /* talking_clear */


/*
* Return 1 if non-local site is talking.
*/
int talking_net(ring_t *r)
{
  int i;
  Audio_member_t m;

  if (r->member_last > 0) return 1;

  for (i = 0; i <= r->member_last; i++) {
    m = r->member[i];
    if (m && m->s->local != m) return 1;
  }
  return 0;
} /* talking_net */
