/*
* Device-independent audio interface, using ahw_...() routines.
*/
#include <tcl.h>
#include <string.h>   /* strcmp(), strchr() */
#include <errno.h>
#include <stdlib.h>   /* atoi() */
#include "ahw.h"
#include "tkaudio.h"
#include "linux_sound.h"

static sound_values_t curr_values;
static int ahw_changed=0;
extern const char *const sys_errlist[];

/*------------------------------------------------------------------------------------------*/
/*
* Variable trace handler for ports and gain.
* Disable if called indirectly from tkaudio_poll.
*/
char *tkaudio_gain_trace(ClientData client, Tcl_Interp *interp,
  char *name1, char *name2, int flags)
{
  char *v = Tcl_GetVar2(interp, name1, name2, flags & TCL_GLOBAL_ONLY);
  int status = 0;
  int value;
  int pr = (long)client;
  char str[40];

  //fprintf(stderr,"tkaudio_gain_trace\n");

  value = atoi(v);
  if (value < 0) {
    value = 0;
    Tcl_SetVar2(interp, name1, name2, "0", flags & TCL_GLOBAL_ONLY);
  }
  else if (value > 100) {
    value = 100;
    Tcl_SetVar2(interp, name1, name2, "100", flags & TCL_GLOBAL_ONLY);
  }


  /* This code has been programmed for handling the SunOS and Solaris AUDIO struct.
     All changes to audioctl are made using this overall audio struct and one function ahw_set.
     In Linux we make different functions for setting gain or port values. */

  /* 0: record, 1: play, 2: monitor */
  /* play */
  if (pr==1) 
    status = audio_set_ovolume(value);
  /* record */
  else if(pr==0)
    status = audio_set_ivolume(value);

  /* 
   * Since the recording and play gain may be quantized at less than
   * 1 unit, we need to read the actual value and set the Tcl
   * variables accordingly.
   */


  if (pr == 0) {
    value = audio_get_ivolume();
    sprintf(str, "%d", value);
    Tcl_SetVar2(interp, "record", "gain", str, TCL_GLOBAL_ONLY);
  }
  else if(pr==1) {
    value = audio_get_ovolume();    
    sprintf(str, "%d", value);
    Tcl_SetVar2(interp, "play", "gain", str, TCL_GLOBAL_ONLY);
  } else {
    Tcl_SetVar2(interp, "monitor", "gain", "0", TCL_GLOBAL_ONLY);
  }


  if (status < 0) return sys_errlist[errno];
  return NULL;
} /* tkaudio_gain_trace */

/*------------------------------------------------------------------------------------------*/
/*
* Tcl variable trace for port.
*/
char *tkaudio_port_trace(ClientData client, Tcl_Interp *interp,
  char *name1, char *name2, int flags)
{
  char *v = Tcl_GetVar2(interp, name1, name2, flags & TCL_GLOBAL_ONLY);
  int pr = (long)client;  /* play or record */
  int status;



  //fprintf(stderr,"tkaudio_port_trace\n");

  /* play */
  if (pr) {
    if (!strcmp(v, "headphone"))
      status = audio_set_oport(AUDIO_HEADPHONE);
    if (!strcmp(v, "line"))
      status = audio_set_oport(AUDIO_LINE_OUT);
  } 
  /* record */
  else {
    if (!strcmp(v, "microphone")) 
      status = audio_set_iport(AUDIO_MICROPHONE);
    if (!strcmp(v, "line")) 
      status = audio_set_iport(AUDIO_LINE_IN);
  }

 
  if (status < 0) return sys_errlist[errno];
  return NULL;
} /* tkaudio_port_trace */

/*------------------------------------------------------------------------------------------*/
/*
* Return list of supported ports: {{microphone line} {speaker
*   headphone}}
* No arguments.
*/
int tkaudio_ports(ClientData client, Tcl_Interp *interp, int argc,
  char *argv[])
{
  int i;
  char** list_rec;
  char** list_play;
  int len_rec, len_play;

  //fprintf(stderr,"tkaudio_ports\n");

  if ((list_rec =  (char**)malloc(10*sizeof(int))) == NULL) {
    perror("malloc");
    return -1;
  }
  if ((list_play =  (char**)malloc(10*sizeof(int))) == NULL) {
    perror("malloc");
    return -1;
  }


  /* get all recording devices */
  if ( (len_rec = audio_get_all_rec_ports(list_rec)) < 0) {
    strcpy(interp->result, sys_errlist[errno]);
    return TCL_ERROR;
  }
  //fprintf(stderr,"len_rec=%d\n",len_rec);
  /* get all playing devices */
  if ( (len_play = audio_get_all_play_ports(list_play)) < 0) {
    strcpy(interp->result, sys_errlist[errno]);
    return TCL_ERROR;
  }
  //fprintf(stderr,"len_play=%d\n",len_play);
  /* append all recording devices to tcl result list */
  Tcl_AppendResult(interp, "{", NULL);
  //fprintf(stderr,"{");
  for (i = 0; i<len_rec; i++) {
    Tcl_AppendElement(interp, list_rec[i]);
    //fprintf(stderr,list_rec[i]);
  }
  Tcl_AppendResult(interp, "} ", NULL);
  //fprintf(stderr, "} ");
  /* append all playing devices to tcl result list */
  Tcl_AppendResult(interp, "{", NULL);
  //fprintf(stderr, "{");
  for (i = 0; i<len_play; i++) {
    Tcl_AppendElement(interp, list_play[i]);
    //fprintf(stderr, list_play[i]);
  }
  Tcl_AppendResult(interp, "} ", NULL);
  //fprintf(stderr,"} ");
  /* that's it */

  free(list_rec);
  free(list_play);

  return TCL_OK;
} /* tkaudio_ports */

/*------------------------------------------------------------------------------------------*/
/*
* Get current values of the audio parameters and set Tcl variables
* accordingly. 
* changes=0: read initial values into Tcl variables.
* changes=1: just read changes.
*/
static void tkaudio_get(ClientData changes)
{
  sound_values_t actual_values;

  char str[40];
  extern Tcl_Interp *Interp;

  //fprintf(stderr,"tkaudio_get\n");

  /* get the actual values from audio hardware */
  actual_values.iport = audio_get_iport();
  actual_values.oport = audio_get_oport();
  actual_values.igain = audio_get_ivolume();
  actual_values.ogain = audio_get_ovolume();
  
  /* input port */
  switch (actual_values.iport) {
  case AUDIO_MICROPHONE :
    if ((actual_values.iport != curr_values.iport) || !changes)
      Tcl_SetVar2(Interp, "record", "port", "microphone", TCL_GLOBAL_ONLY);
    break;
  case AUDIO_LINE_IN :
    if ((actual_values.iport != curr_values.iport) || !changes)
      Tcl_SetVar2(Interp, "record", "port", "line", TCL_GLOBAL_ONLY);
    break;
  default:
    Tcl_SetVar2(Interp, "record", "port", "invalid", TCL_GLOBAL_ONLY);
  }

  /* output port */
  switch (actual_values.oport) {
  case AUDIO_HEADPHONE :
    if ((actual_values.oport != curr_values.oport) || !changes)
      Tcl_SetVar2(Interp, "play", "port", "headphone", TCL_GLOBAL_ONLY);
    break;
  case AUDIO_LINE_OUT :
    if ((actual_values.oport != curr_values.oport) || !changes)
      Tcl_SetVar2(Interp, "play", "port", "line", TCL_GLOBAL_ONLY);
    break;
  default:
    Tcl_SetVar2(Interp, "play", "port", "invalid", TCL_GLOBAL_ONLY);
  }

  /* input gain */
  if ((actual_values.igain != curr_values.igain) || !changes) {
    sprintf(str, "%d", actual_values.igain);
    Tcl_SetVar2(Interp, "record", "gain", str, TCL_GLOBAL_ONLY);
  }
 
  /* output gain */
  if ((actual_values.ogain != curr_values.ogain) || !changes) {
    sprintf(str, "%d", actual_values.ogain);
    Tcl_SetVar2(Interp, "play", "gain", str, TCL_GLOBAL_ONLY);
  }

  Tcl_SetVar2(Interp, "monitor", "gain", "0", TCL_GLOBAL_ONLY);
  
  /* I don't want to change the interface to PMM and TCL. Therefore 
     this values are sent to the upper layer. The Linux mixer does not
     provide any information about sample_rate, errors, encoding etc. */
  Tcl_SetVar2(Interp, "record", "samples", "invalid", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(Interp, "play", "samples", "invalid", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(Interp, "record", "active", "invalid", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(Interp, "play", "active", "invalid", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(Interp, "record", "errors", "invalid", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(Interp, "play", "errors", "invalid", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(Interp, "record", "encoding", "invalid", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(Interp, "play", "encoding", "invalid", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(Interp, "record", "channels", "invalid", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(Interp, "play", "channels", "invalid", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(Interp, "record", "sample_rate", "invalid", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(Interp, "play", "sample_rate", "invalid", TCL_GLOBAL_ONLY);

  memcpy(&curr_values, &actual_values, sizeof(sound_values_t));

  ahw_changed = 1;

} /* tkaudio_get */

/*------------------------------------------------------------------------------------------*/
/*
* Signal handler invoked if audio parameter changes.  Since this may
* trigger lots of Tcl and Tk code, and doing this from within a signal
* handler leads to X crashing and other random effects, we only
* schedule a delayed call from the signal, hoping that that is safe.
*/
void tkaudio_poll(int sig)
{ 
  //fprintf(stderr,"tkaudio_poll\n");
   Tcl_CreateTimerHandler(50, tkaudio_get, (ClientData)1); 
} /* tkaudio_poll */


/*------------------------------------------------------------------------------------------*/
int tkaudio_state(ClientData client, Tcl_Interp *interp, int argc,
  char *argv[])
{
  //fprintf(stderr,"tkaudio_state\n");
  tkaudio_get((ClientData)1);
  return TCL_OK;
} /* tkaudio_state */


/*------------------------------------------------------------------------------------------*/
/*
* Set up a periodic timer that checks the audio device for any
* changes. For systems that signal changes, ahw_get() is only
* called if the ahw_changed flag is set.
*/
void tkaudio_timer(ClientData client)
{
  //fprintf(stderr,"tkaudio_timer\n");
  /* get changed or initial values */
  if (ahw_changed || client == 0) tkaudio_get((ClientData)ahw_changed);
  /* schedule next timer */
  Tcl_CreateTimerHandler(1000, tkaudio_timer, (ClientData)1);
} /* tkaudio_timer */
