/*
sgi_audio  --  audio interface for Silicon graphics

Audio interface for the high-quality digital audio subsystems on IRIS
4D/30, 4D/35, and Indigo systems, and future IRIS audio subsystems.
Note that the IRIS 4D/20 has a different (/dev/audio) interface.

The device can only switch input ports, not output ports.

DC offset can be significant (1300 or more).

Recording/playback gain: 0..255

Copyright 1994 by Henning Schulzrinne; all rights reserved.
*/

#include <audio.h>
#include <unistd.h>        /* fork() */
#include <memory.h>        /* memcmp() */
#include "audio_generic.h"
#include "ahw.h"
#include "util.h"
#include "config.h"

static char rcsid[] = "$Id: sgi_audio.c,v 1.1.1.1 1998/04/21 17:05:50 sho Exp $";

#define ALPHA 0.009437412

static struct ALc {
  ALport port;
  ALconfig config;
  audio_descr_t ad;  /* current configuration */
  int fillpoint;     /* current queue size */
  int samples;       /* samples read and written */
  int channels;
  int byte_sample;   /* bits to shift to convert bytes to samples */
} ALplay, ALrecord;

int ahw_changed = 1;

/*
* Create configuration datastructure. Return 0 on success, -1 on error.
*/
static int ahw_config(void)
{
  if (!ALplay.config)   ALplay.config   = ALnewconfig();
  if (!ALrecord.config) ALrecord.config = ALnewconfig();  
  if (ALplay.config && ALrecord.config) return 0;
  return -1;
} /* ahw_config */

/*
* Open audio hardware.  Return file descriptor number.  Also sets
* maximum fillpoint for input and output audio queues.  Return -1 on
* error.
*/
int ahw_open(int ndelay)
{
  if (ahw_config() < 0) return -1;
  if (!ALrecord.port) {
    ALsetqueuesize(ALrecord.config, 8000);
    ALrecord.port = ALopenport("nevot-rec", "r", ALrecord.config);
    ALsetfillpoint(ALrecord.port, ALrecord.fillpoint);
  }
  if (!ALplay.port) {
    ALsetqueuesize(ALplay.config, 8000);
    ALplay.port = ALopenport("nevot-play", "w", ALplay.config);
  }

  return ALgetfd(ALrecord.port);
} /* ahw_open */

/*
* Close audio hardware. Return integer file descriptor for input device.
*/
int ahw_close(void)
{
  int au = -1;

  if (ALrecord.port) {
    au = ALgetfd(ALrecord.port);
    ALcloseport(ALrecord.port);
    ALrecord.port = 0;
  }
  if (ALplay.port) {
    ALcloseport(ALplay.port);
    ALplay.port = 0;
  }
  return au;
} /* ahw_close */


static int enc2bytes(audio_encoding_t enc)
{
  if      (enc == AE_L8)  return 1;
  else if (enc == AE_L16) return 2;
  else if (enc == AE_L24) return 4;
  else if (enc == AE_L32) return 4;
  return 0;
} /* enc2bytes */

/*
* Set audio device characteristics (port, gain).
* Return -1 on error, 0 on success.
*/
int ahw_set(int pr, audio_set_t p, int value)
{
  audio_state_t as;
  static long atten[] = { 34, 37, 40, 45, 50, 55, 63 };
  struct ALc *alc;
  int port;             /* port value */
  long pv_buf[10];      /* parameter/value buffer */
  int c = 0;            /* counter */
  int i, j;

#define ALPAR(f, v) {pv_buf[c++] = (f); pv_buf[c++] = (v);}

  alc = pr ? &ALplay : &ALrecord; 
  if (ahw_config() < 0) return -1;
  switch(p) {
  case AS_fillpoint:
    alc->fillpoint = value;
    break;

  case AS_gain:
    if (pr) {
      value = (long)(pow(10.0, (double)value * ALPHA) + 0.5);
      ALPAR(AL_LEFT_SPEAKER_GAIN, value);
      ALPAR(AL_RIGHT_SPEAKER_GAIN, value);
    }
    else {
      value = (255 - value)/255.0 * 39.0 + 0.5;
      if (value > 32) value = atten[value-33];
      value <<= 2;
      ALPAR(AL_LEFT_INPUT_ATTEN,  value);
      ALPAR(AL_RIGHT_INPUT_ATTEN, value);
    }
    break;

  case AS_gain_step:
    ahw_get(&as);
    ahw_set(pr, AS_gain, (pr ? as.play.gain[0] : as.record.gain[0]) + value);
    break;

  case AS_port:
    if (pr == 0) {
      if      (value & AP_in_line)       port = AL_INPUT_LINE;
      else if (value & AP_in_microphone) port = AL_INPUT_MIC;
      else if (value & AP_in_digital)    port = AL_INPUT_DIGITAL;
      else return -1;
      ALPAR(AL_INPUT_SOURCE, port);
    }
    break;
  }

  if (c > 0) ALsetparams(AL_DEFAULT_DEVICE, pv_buf, c);
  return 0;
} /* ahw_set */

/*
* Set audio input and output format (encoding, sampling rate,
* channels).  SGI can handle different sampling rates for input and
* output, but only linear input/output.  We currently do not handle
* 24-bit or 8-bit linear. Return -1 on error, 0 on success.
*/
int ahw_fmt(audio_descr_t *record, audio_descr_t *play)
{
  int pr;
  struct ALc *alc;
  long pv_buf[10];      /* parameter/value buffer */
  int c = 0;            /* counter */
  audio_encoding_t enc; /* encoding */

  /* sample width */
  record->encoding = AE_L16;
  play->encoding   = AE_L16;

  /* Compare to existing values to see if we need to do anything. */
  if (memcmp(record, &ALrecord.ad, sizeof(audio_descr_t)) == 0 &&
      memcmp(play,   &ALplay.ad,   sizeof(audio_descr_t)) == 0) return 0;

  if (ahw_config() < 0) return -1;

  for (pr = 0; pr < 2; pr++) {
    enc = pr ? play->encoding : record->encoding;
    alc = pr ? &ALplay : &ALrecord;
    if (ALsetwidth(alc->config, enc2bytes(enc)) < 0) return -1;
    if      (ALgetwidth(alc->config) == AL_SAMPLE_8)  alc->byte_sample = 0;
    else if (ALgetwidth(alc->config) == AL_SAMPLE_16) alc->byte_sample = 1;  
    else if (ALgetwidth(alc->config) == AL_SAMPLE_24) alc->byte_sample = 2;  
  }
  ALPAR(AL_OUTPUT_RATE, play->sample_rate);
  ALPAR(AL_INPUT_RATE,  record->sample_rate);

  /* channels */
  ALplay.channels = play->channels;
  ALrecord.channels = record->channels;
  if (ALsetchannels(ALplay.config, play->channels) < 0) return -1;
  if (ALsetchannels(ALrecord.config, record->channels) < 0) return -1;

  /* sample rate */
  if (c > 0) {
    if (ALsetparams(AL_DEFAULT_DEVICE, pv_buf, c) < 0) return -1;
  } 

  return 0;
} /* ahw_fmt */

/*
* Get current audio hardware configuration.  Return 0 on success, -1
* on error.
*/
int ahw_get(audio_state_t *as)
{
  int port, x;
  int c;
  long pv_buf[50];
  int i, k;
  static short atten[] = 
    {                33, 33, 33, 34, 34, 35, /* 34 .. 39 */
     35, 35, 35, 36, 36, 36, 36, 36, 36, 36, /* 40 .. 49 */     
     37, 37, 37, 38, 38, 38, 38, 38, 39, 39, /* 50 .. 59 */
     39, 39, 39, 39};                        /* 60 .. 63 */

  if (ahw_config() < 0) return -1;
  c = ALqueryparams(AL_DEFAULT_DEVICE, pv_buf, sizeof(pv_buf)/sizeof(pv_buf[0]));
  if (c < 0) return -1;
  if (ALgetparams(AL_DEFAULT_DEVICE, pv_buf, c) < 0) return -1;
  for (k = 0; k < c; k += 2) {
    i = 0;
    switch(pv_buf[k]) {
    case AL_RIGHT_SPEAKER_GAIN:
      i = 1;
    case AL_LEFT_SPEAKER_GAIN:
      as->play.gain[i] = log10(pv_buf[k+1]) / ALPHA + 0.5;
      break;

    case AL_RIGHT_INPUT_ATTEN:
      i = 1;
    case AL_LEFT_INPUT_ATTEN:
      x = pv_buf[k+1] >> 2;
      if (x >= 34) x = atten[x - 34];
      as->record.gain[i] = 255 - x/39. * 255. + 0.5;
      break;

    case AL_RIGHT_MONITOR_ATTEN:
      i = 1;
      break;
    case AL_LEFT_MONITOR_ATTEN:
      as->monitor_gain[i] = 255 - x/39. * 255. + 0.5;
      break;

    case AL_INPUT_SOURCE:
      port = pv_buf[k+1];
      if (port == AL_INPUT_MIC)          as->record.port = AP_in_microphone;
      else if (port == AL_INPUT_LINE)    as->record.port = AP_in_line;
      else if (port == AL_INPUT_DIGITAL) as->record.port = AP_in_digital; 
      break;

    case AL_INPUT_RATE:
      as->record.ad.sample_rate = pv_buf[k+1];
      break;

    case AL_OUTPUT_RATE:
      as->play.ad.sample_rate   = pv_buf[k+1];
      break;
    }
  }
  as->play.port = AP_out_speaker;

  /* Error status. */
  pv_buf[0] = AL_ERROR_NUMBER;
  pv_buf[1] = 0;
  pv_buf[2] = AL_ERROR_TYPE;
  pv_buf[3] = 0;

  for (i = 0; i < 2; i++) {
    struct ALc *alc;
    audio_ws_t *ws;

    alc = (i == 0) ? &ALrecord   : &ALplay;
    ws  = (i == 0) ? &as->record : &as->play;

    ws->ad.encoding         = AE_L16;
    ws->ad.channels         = ALgetchannels(alc->config);
    ws->waiting             = 0;
    ws->active              = !!alc->port;
    ws->fillpoint           = alc->fillpoint;
    ws->samples             = alc->samples;
    if (alc->port) ALgetstatus(alc->port, pv_buf, 4);
    ws->errors              = pv_buf[1];
  }

  return 0;
} /* ahw_get */

/*
* This function is undefined for SGI.
*/
int ahw_waiting(void)
{
  return 0;
} /* ahw_waiting */


/*
* SGI does not support tracking.
*/
void ahw_track(signal_func_t func)
{
} /* ahw_track */


/*
* Return number of samples in play queue, <0 if underflow.  If the
* number is negative, an underflow has occurred and we also set
* 'underflow'.
*/
int ahw_write_queue(int *underflow)
{
  int q;
  q = ALgetfilled(ALplay.port);
  *underflow = (q < 0);
  return q;
} /* ahw_write_queue */


/*
* Return number of samples waiting to be read.
* Return <0 if error.
*/
int ahw_read_queue(int *overflow)
{
  *overflow = 0;
  return ALgetfilled(ALrecord.port);
} /* ahw_read_queue */

/*
* Write audio buffer to audio device.
* Since we don't select on the audio output device, ALsetfillpoint
* shouldn't be needed.
*/
int ahw_write(char *buffer, int bytes)
{
  register int samplecount = bytes >> ALplay.byte_sample;

  if (bytes == 0) return 0;
  ALwritesamps(ALplay.port, buffer, samplecount);
  ALplay.samples += samplecount;
  return bytes;
} /* ahw_write */

/*
* Read audio frames from audio device.  Always read even number of
* samples to avoid difficulties with stereo.  man page says that
* "ALreadsamps and Alwritesamps may change the fillpoint so you should
* set it just before you call select or poll".  The fillpoint is set
* to the number of samples expected (see ahw_config(AS_fillpoint)). 
*/
int ahw_read(char *buffer, int bytes)
{
  int samplecount = bytes >> ALrecord.byte_sample;
  int available   = ALgetfilled(ALrecord.port);

  if (samplecount > available) samplecount = available;
  ALreadsamps(ALrecord.port, buffer, samplecount);
  ALsetfillpoint(ALrecord.port, ALrecord.fillpoint);
  ALrecord.samples += samplecount;
  return samplecount << ALrecord.byte_sample;
} /* ahw_read */
