/***********************************************************************************
 *
 * procedures which perform manipulation and searching in strings, requests and 
 * responses
 *
 * partially based on cd by Christian Zahl
 *
 * Copyright 1997 by Stefan Hoffmann
 * GMD Fokus
 *
 ***********************************************************************************/



#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include "debug.h"




/* str_tolower ***********************************************************************
 *
 * convert string to lower case
 *
 *  in:
 *  - s_normal: string which sould be converted to lower case
 *
 *  returns:
 *  - lower case string
 *
 ***********************************************************************************/
char *str_tolower (char *s_normal)
{
  static char     func[] = "str_tolower";
  char            *s_lower,*s;


  DEBUG (4, syslog (LOG_DEBUG, "in %s",func);)

  s_lower = (char *)malloc (strlen(s_normal)+1);
  s = s_lower;
  while ((*(s++) = tolower(*(s_normal++))) != '\0');
  return s_lower;
} /* str_tolower);



/* str_search ***********************************************************************
 *
 *  locate  first occurrence of string s2 in string s1, not case sensitive
 *
 *  in:
 *  - s1: pointer to string in which s1 should be searched
 *  - s2: pointer to sting which should be searched
 * 
 *  returns: 
 *  - pointer to the located string in s1
 *  - NULL, if no match occurrs
 *
 ***********************************************************************************/
char *str_search(char *s1, char *s2)
{
  static char     func[] = "str_search";
  char            *s1_lower, *s2_lower, *s, *s1_orig;
  int             i;


  DEBUG (4, syslog (LOG_DEBUG, "%s",func);)

  
  s1_orig = s1;
  /* switch both strings to lower case */
  s1_lower = str_tolower(s1);
  s2_lower = str_tolower(s2);

  if ((s = strstr(s1_lower, s2_lower)) != NULL)
    s = s1_orig + (s - s1_lower);
  free (s1_lower);
  free (s2_lower);
  return s;
} /* str_search */



/* str_cmp *************************************************************************
 *
 *  compares string s1 and string s2, not case sensitive
 *
 *  in: 
 *  - s1, s2: pointers to strings which should be compared
 *
 *  returns:
 *  - int < 0, if s1 < s2
 *  - int = 0, if s1 = s2
 *  - int > 0, if s1 > s2
 *
 ***********************************************************************************/
int str_cmp (char *s1, char *s2)
{
  static char     func[] = "str_cmp";
  char            *s1_lower, *s2_lower;
  int             i;
  

  DEBUG (4, syslog (LOG_DEBUG, "%s",func);)
    
  /* switch both strings to lower case */
  s1_lower = str_tolower(s1);
  s2_lower = str_tolower(s2);

  i = strcmp(s1_lower, s2_lower);

  free (s1_lower);
  free (s2_lower);
  return i;
} /* str_cmp */



/* skip_white_spaces ***************************************************************
 *
 *  remove leading SPACE and TAB charcters of a string
 *
 *  in:
 *  - p: string
 *  
 *  returns:
 *  - string without leading white spaces
 *
 ***********************************************************************************/
char *skip_white_spaces (char *p)
{
  DEBUG (4, syslog (LOG_DEBUG, "skip_white_spaces");)

  /* skip any leading whitespaces */
  while (1) {
    if (*p == '\0')
      break;
    if ((*p != ' ') && (*p != '\t'))
      break;
    p++;
  }
  return p;
} /* skip_white_spaces */



/* get_field **************************************************************************
 *
 *  get next field of a string
 *  fileds are separated by SPACE or TAB characters
 *
 *  in:
 *  - p: pointer to string in which next field should be extracted
 *  - str: pointer to string in which extraced field will be stored
 *  - len: length or string str
 *  
 *  returns:
 *  - pointer to the end of the extracted field in p
 *  - NULL on failure
 *
 ***********************************************************************************/
char *get_field (char *p, char *str, int len)
{
 
  DEBUG (4, syslog (LOG_DEBUG, "get_field");)
 
  /* skip any leading whitespaces */
  p = skip_white_spaces (p);
    
  /* read up to next whitespace or eol */
  while (len-- > 0) {
    if (*p == '\0')
      break;
    if (*p == ' ')
      break;
    if (*p == '\t')
      break;
    if (*p == '\n')
      break;
    if (*p == '\r')
      break;
    *str++ = *p++;
  }
  if (len < 0)
    return NULL;
  *str = '\0';
  return p;
} /* get_field */



/* get_line **************************************************************************
 *
 *  get string beginning at p up to the end of line
 *
 *  in:
 *  - p: pointer to the beginning of the line
 *  - str: pointer to string in which line will be stored
 *  - len: length or string str
 *  
 *  returns:
 *  - pointer to the end of the extracted line
 *  - NULL on failure
 *
 ***********************************************************************************/
char *get_line (char *p, char *str, int len)
{
 
  DEBUG (4, syslog (LOG_DEBUG, "get_line");)
 
  /* skip any leading whitespaces */
  p = skip_white_spaces (p);
    
  /* read up to eol */
  while (len-- > 0) {
    if (*p == '\0')
      break;
    if (*p == '\n')
      break;
    if (*p == '\r')
      break;
    *str++ = *p++;
  }
  if (len < 0)
    return NULL;
  *str = '\0';
  return p;
} /* get_line */



/* search_SIP_header **************************************************************
 * 
 *  search for a SIP header entry in msg specified by header_name
 *
 *  in:  
 *  - msg: pointer to message
 *  - header_name: pointer to the header name
 *  - short_name: pointer to the short header name
 *   
 *  returns:
 *  - pointer to the field 
 *  - NULL: header not found
 *
 ***********************************************************************************/
char *search_SIP_header (char *msg, char *header_name, char *short_name)
{
  static char     func[] = "search_SIP_header";
  char            *p;
  

  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  /***********************************************************************************
   * search for the specified header in msg 
   ***********************************************************************************/
  if ((p = str_search(msg, header_name)) == NULL) {
    if ((p = str_search(msg, short_name)) == NULL) 
      return NULL;
    else 
      p += strlen(short_name);
  }
  else
    p += strlen(header_name);
     
  /***********************************************************************************
   * check if header was found in the SIP header
   ***********************************************************************************/
  if (p > strstr(msg, "\r\n\r\n"))
    return NULL;
  return p;
} /* search_SIP_header */



/* delete_SIP_line ******************************************************************
 * 
 *  delete a line in a SIP request or a response header
 *
 *  in:  
 *  - s: pointer to line which should be deleted
 *
 *
 ***********************************************************************************/
int delete_SIP_line (char *s)
{
  char     *p;
 
 
  DEBUG (4, syslog (LOG_DEBUG, "delete_SIP_line");)
    
  /* find end of line */
  if ((p = strchr(s, '\n')) == NULL) 
    return (-1);
  memmove(s, p+1, strlen(p));
  return 0;
} /* delete_SIP_line */



/* insert_SIP_line ******************************************************************
 * 
 *  insert a line in a SIP request or response header
 *
 *  in:
 *  - p: pointer to place where line should be inserted
 *  - s: pointer to string which should be inserted
 *
 ***********************************************************************************/
void insert_SIP_line (char *p, char *s)
{
 
  DEBUG (4, syslog (LOG_DEBUG, "insert_SIP_line");)

  memmove (p+strlen(s)+2, p, strlen(p)+1);
  memmove (p, s, strlen(s));

  /* terminate inserted line by CRLF */
  p[strlen(s)] = '\r';
  p[strlen(s)+1] = '\n';

  return;
} /* insert_SIP_line */



