/***********************************************************************************
 *
 * functions to 
 *  - deceide whether to forward requets or to send an Alternative Address reply
 *  - perform forwarding
 *
 * Copyright 1997 by Stefan Hoffmann
 * GMD Fokus
 *
 ***********************************************************************************/



#include "debug.h"
#include <syslog.h>
#include "misc.h"
#include <string.h>
#include <stdio.h>
#include "udp.h"
#include "timer.h"
#include "list.h"
#include "response.h"
#include "notify.h"
#include "multimer.h"
#include "defines.h"
#include "isc.h"
#include "event.h"
#include "reply_codes.h"



extern char myAddr[17];
extern int isc_connections;
extern int isc_socket;
extern int udp_server_socket;
extern int sip_port;



/* get_via_header_no ***************************************************************
 *
 *  - calculate number of via headers in a request packet
 *
 *  in:
 *  - req: pointer to SIP request
 *  
 *  returns:
 *  - number of via headers
 *
 ***********************************************************************************/
int get_via_header_no (char *req)
{
  static char     func[] = "get_via_header_no";
  char            *p;
  int             i = 0;


  DEBUG (2, syslog (LOG_DEBUG, "%s", func);)
     
  p = req;
  while ((p = str_search(p, "\nv")) != NULL) {
    if (p > strstr(req, "\r\n\r\n")) 
      break;
    p++;
    i++;
  }
  return i;
} /* get_via_header_no */



/* forward_request ******************************************************************
 *
 *  forward a SIP request to next SIP server or via local multicast to isc
 *
 *  in:
 *  - loc: pointer to entry in locations_list
 *  
 ***********************************************************************************/
void forward_request(loc_list_t *loc)
{
  static char     func[] = "forward_request";
  inv_list_t      *inv; /* pointer to invitation_list entry */
  char            new_line[MAX_LINE_LENGTH];
  char            *p;
  int             sec;
  struct timeval  time;
  struct timeval  isc_time;
  struct timeval  req_time;
  struct timeval  retransmit_time;
    

  DEBUG (2, syslog (LOG_DEBUG, "%s", func);)
   
  /* get invitation_list entry belonging to loc */
  inv = loc->inv_parent;
  /* create new 1. request line */
  delete_SIP_line(inv->request);
  sprintf(new_line, "INVITE sip://%s@%s %s",loc->user, loc->addr, SIP_VERSION);
  insert_SIP_line(inv->request, new_line);

  /***********************************************************************************
   * if request should be forwarded to the local host (myAddr) it will be send via local 
   * multicast (isc_socket) to isc to notify the called user 
   ***********************************************************************************/
  if (!(strcmp(loc->addr, myAddr))) {
    /***********************************************************************************
     * forward request to isc 
     ***********************************************************************************/
    /* check if there is already a connection (socket) from the daemon to isc */      
    if (isc_connections <= 0) {
      /* create local multicast socket to isc */ 
      if ((isc_socket = UDP_connect(LOCAL_MCAST_ADDR, LOCAL_MCAST_PORT, 0)) == -1) {
        /***********************************************************************************
         * creation of a socket to isc failed; handle like no response from isc was received 
         * if there are still locations to query, otherwise send error reply 
         ***********************************************************************************/
        if (inv->locations > 1)
          isc_request_timeout((Notify_client)inv);
        else {
          inv->locations = 0;
          reply (inv, SERVER_INTERNAL_ERROR, "Server Internal Error");
        }
        return;
      }
      /* install event handler for new local multicast socket */
      notify_set_input_func(0, (Notify_func_input)isc_socket_event, isc_socket);
    }
    ++isc_connections;

    /* send request */
    if (send_to_isc(inv, inv->request) < 0) {
      /***********************************************************************************
       * sending the request to isc failed; handle like no response from isc was received 
       * if there are still locations to query, otherwise send error reply 
       ***********************************************************************************/
      if (inv->locations > 1)
        isc_request_timeout((Notify_client)inv);
      else {
        inv->locations = 0;
        reply (inv, SERVER_INTERNAL_ERROR, "Server Internal Error"); 
      }
      return;
    }
    /***********************************************************************************
     * start T_isc timer for receiving a "TRYING" answer form isc to check if isc 
     * received the request 
     ***********************************************************************************/
    time.tv_sec = 2;
    time.tv_usec = 0;
    timer_set(&time, (Notify_func)isc_timeout, (Notify_client)inv, 1); 
  }
  else {
    /***********************************************************************************
     * the request should be forwarded to another host (loc->addr != myAddr)
     * forward request to next sip server using TCP if no parallel search is performed 
     * and if request was received via TCP, otherwise use UDP
     ***********************************************************************************/
    if ((inv->state != PARALLEL_SEARCH) && (inv->transport_to_client == SOCK_STREAM)) {
      /***********************************************************************************
       * forward via TCP
       ***********************************************************************************/
      /* create new via header */
      sprintf(new_line, "v:%s/TCP %s", SIP_VERSION, myAddr);
   
      /* insert via header before existing ones */
      p = str_search(inv->request, "\nv")+1;
      insert_SIP_line(p, new_line);
   
      /* existing connection to next SIP server ? */
      if (inv->socket_to_server == 0) {
        /* create connection to next server */
        if((inv->socket_to_server = TCP_connect(loc)) <= 0) {
          /* decrease counter of outstanding responses */
          --inv->locations;
          reply(inv, SERVER_INTERNAL_ERROR, "Server Internal Error");       
          delete_SIP_line(p);
          return;
        }
        inv->transport_to_server = SOCK_STREAM;
        /* install event handler for socket_to_client */
        notify_set_input_func((Notify_client)inv, (Notify_func_input)socket_event, inv->socket_to_server);
      }
      
      if (TCP_send(inv->socket_to_server, inv->request) < 0) {
        /* decrease counter of outstanding responses */
        -- inv->locations;
        reply(inv, SERVER_INTERNAL_ERROR, "Server Internal Error");       
        delete_SIP_line(p);
        return;
      }
      /* rebuild the original incoming request */
      delete_SIP_line(p);
   
      /***********************************************************************************
       * start T_request timer for receiving a final response from the next server 
       ***********************************************************************************/
      sec = MAX_TIMEOUT - get_via_header_no(inv->request) * 10;
      req_time.tv_sec = sec;
      req_time.tv_usec = 0; 
      timer_set(&req_time, (Notify_func)request_timeout, (Notify_client)inv, 1);    
    }
    else {
      /***********************************************************************************
       * forward via UDP 
       ***********************************************************************************/
      /* create via header */
      sprintf(new_line, "v:%s/UDP %s", SIP_VERSION, myAddr);

      /* insert via header before existing ones */
      p = str_search(inv->request, "\nv")+1;
      insert_SIP_line(p, new_line);

      if (UDP_send (loc->addr, sip_port, udp_server_socket, 1, inv->request) < 0) {
        /* decrease counter of outstanding responses and set resonse_received flag for this location*/
        --inv->locations;
        loc->response_received = TRUE;
        reply(inv, SERVER_INTERNAL_ERROR, "Server Internal Error");       
        delete_SIP_line(p);
        return;
      }
   
      /* rebuild the original incoming request */
      delete_SIP_line(p);
   
      /***********************************************************************************
       * start T_retransmit timer for retransmitting the request until a final response was received 
       ***********************************************************************************/
        /* start T_retransmit timer */
        retransmit_time.tv_sec = RETRANSMIT_TIME;
        retransmit_time.tv_usec = 0;
        timer_set(&retransmit_time, (Notify_func)retransmit_timeout, (Notify_client)loc, 1); 
    }
  }
} /* forward_request */
 


/* proxy ****************************************************************************
 *
 *  invoked if the invited user is possibly located at several locations
 *  makes a parallel search if  this is the first proxy  or sends an ALTERNATIVE 
 *  ADDRESS reply
 *
 *  in:
 *  - inv: pointer to entry in invitation_list
 *  
 ***********************************************************************************/
void proxy (inv_list_t *inv)
{
  static char     func[] = "proxy";
  loc_list_t      *loc;
  char            *p;


  DEBUG (2, syslog (LOG_DEBUG, "%s", func);)
    
  /***********************************************************************************
   * check if this is the first proxy
   ***********************************************************************************/
  if (get_via_header_no(inv->request) <= 1) {
    /***********************************************************************************
     * first proxy 
     * forward request to all locations 
     ***********************************************************************************/
    inv->state = PARALLEL_SEARCH;
    for (loc = inv->location_list; loc != NULL; loc = loc->next) { 
      forward_request(loc); 
    }
  }
  else {
    /***********************************************************************************
     * not the first proxy 
     * if user is possibly located at this host, send request to isc; if isc doen't 
     * reply, send an ALTERNATIVE ADDRESS reply with the other locations to the client
     ***********************************************************************************/ 
    if ((loc = loc_list_search(inv, myAddr)) != NULL) 
      forward_request(loc);
    else {
      /***********************************************************************************
       * send ALTERNATIVE ADDRESS reply 
       ***********************************************************************************/
      inv->locations = 0;   /* make sure that check_response_priority will send the reply */
      reply(inv, ALTERNATIVE_ADDRESS, "Alternative Address");
    }
  }
} /* proxy */
 
  
